from __future__ import annotations

import os
import re
import sys
import tempfile
import time
import warnings
from inspect import stack


def remove_regex_capture_group_names(regex):
    # PostgreSQL will choke on named capture groups
    return re.sub(r'\?<\w+>', '', regex)


def pythonize_regex_capture_group_names(regex):
    # Unlike standard regex syntax, Python capture groups have a capital P:  (?P<group_name>.*?)
    return re.sub(r'\?<(\w+)>', r'?P<\1>', regex)


def replace_tokens_in_regex(regex, group_dict, escape=True):
    for name, value in group_dict.items():
        regex = regex.replace('${%s}' % name, re.escape(value) if escape else value)

    return regex


def os_lock(name, target, args=(), kwargs=None, timeout=7.0, retry_period=0.1):
    """
    Adapted from https://github.com/ses4j/backup/blob/master/oslockedaction.py
    To use, just wrap your python call with an os_lock() call.  Instead of:
    >>> time.sleep(5.0)
    Do:
    >>> os_lock(name="mylock", target=time.sleep, args=(5.0,))
    Nobody else using the same lockfile will be able to run at the same time.
    If timeout is None, it will never timeout if the lockfile exists.
    Otherwise it will just try to remove the lockfile after "timeout" seconds and do 'action()' anyway.
    retry_period is a float in seconds to wait between retries.
    """
    lockfile = os.path.abspath(os.path.join(tempfile.tempdir, f'{name}.lock'))

    f = None
    started_at = time.time()
    kwargs = dict() if kwargs is None else kwargs

    try:
        while True:
            if timeout is not None and time.time() > started_at + timeout:
                # timed out, it must be stale... (i hope)  so remove it.

                if time.time() > (started_at + (timeout * 2)):
                    raise RuntimeError(f"FAIL: os_lock() can't acquire directory lock, because {lockfile} was not "
                                       "cleaned up. Delete it manually if the process is finished.")

                try:
                    os.remove(lockfile)
                except OSError:
                    time.sleep(retry_period)
                    continue

            try:
                f = os.open(lockfile, os.O_RDWR | os.O_CREAT | os.O_EXCL)
                os.write(f, bytes(f'{str(__file__)} made this, called from {sys.argv[0]}', encoding='utf-8'))
                break
            except OSError:
                time.sleep(retry_period)

        target(*args, **kwargs)
    finally:
        if f:
            os.close(f)
            os.remove(lockfile)


def deprecation_warning(message, *, stack_index=2):
    frame_info = stack()
    if len(frame_info) > stack_index:
        warnings.warn(
            f'{message} (from line {frame_info[stack_index].lineno} in file "{frame_info[stack_index].filename}")')
    else:
        warnings.warn(message)


def get_platform():
    import platform
    plat = platform.platform()
    if plat.lower().find('windows') != -1:
        return 'windows'
    if plat.lower().find('linux') != -1:
        return 'linux'
    if plat.lower().find('darwin') != -1 or plat.lower().find('macos') != -1:
        return 'osx'


def is_windows():
    return get_platform() == 'windows'


def is_linux():
    return get_platform() == 'linux'


def is_osx():
    return get_platform() == 'osx'


def is_osx_aarch64():
    import platform
    return get_platform() == 'osx' and platform.platform().lower().find('arm64') != -1


def get_home_dir(user=None):
    if is_windows():
        home = os.environ['USERPROFILE']
    else:
        user_to_expand = user if user else ''
        home = os.path.expanduser(f'~{user_to_expand}')

    return home


def get_test_with_file():
    return os.path.normpath(
        os.path.join(os.path.dirname(__file__), '..', 'sdk', 'test-with.txt'))


def get_test_with_root_dir():
    test_with_file = get_test_with_file()
    test_with = os.path.normpath(os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'))
    if os.path.exists(test_with_file):
        with open(test_with_file, 'r') as f:
            lines = f.readlines()
            test_with = lines[0]

    return test_with


def do_with_retry(func, timeout_sec=5, retry_sleep_sec=0.1):
    import time
    timer = time.time()
    while True:
        # noinspection PyBroadException
        try:
            func()
            break
        except BaseException:
            if time.time() - timer > timeout_sec:
                raise

            time.sleep(retry_sleep_sec)


def cleanse_filename(filename, replacement_char='_'):
    return re.sub(r'[:"%\\/<>^|?*&\[\]]', replacement_char, filename)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.3.2-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class RegressionOutputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'adjusted_r_squared': 'float',
        'error_sum_squares': 'float',
        'intercept': 'float',
        'intercept_standard_error': 'float',
        'is_uncertain': 'bool',
        'r_squared': 'float',
        'regression_sum_squares': 'float',
        'suggested_p_value_cutoff': 'float'
    }

    attribute_map = {
        'adjusted_r_squared': 'adjustedRSquared',
        'error_sum_squares': 'errorSumSquares',
        'intercept': 'intercept',
        'intercept_standard_error': 'interceptStandardError',
        'is_uncertain': 'isUncertain',
        'r_squared': 'rSquared',
        'regression_sum_squares': 'regressionSumSquares',
        'suggested_p_value_cutoff': 'suggestedPValueCutoff'
    }

    def __init__(self, adjusted_r_squared=None, error_sum_squares=None, intercept=None, intercept_standard_error=None, is_uncertain=False, r_squared=None, regression_sum_squares=None, suggested_p_value_cutoff=None):
        """
        RegressionOutputV1 - a model defined in Swagger
        """

        self._adjusted_r_squared = None
        self._error_sum_squares = None
        self._intercept = None
        self._intercept_standard_error = None
        self._is_uncertain = None
        self._r_squared = None
        self._regression_sum_squares = None
        self._suggested_p_value_cutoff = None

        if adjusted_r_squared is not None:
          self.adjusted_r_squared = adjusted_r_squared
        if error_sum_squares is not None:
          self.error_sum_squares = error_sum_squares
        if intercept is not None:
          self.intercept = intercept
        if intercept_standard_error is not None:
          self.intercept_standard_error = intercept_standard_error
        if is_uncertain is not None:
          self.is_uncertain = is_uncertain
        if r_squared is not None:
          self.r_squared = r_squared
        if regression_sum_squares is not None:
          self.regression_sum_squares = regression_sum_squares
        if suggested_p_value_cutoff is not None:
          self.suggested_p_value_cutoff = suggested_p_value_cutoff

    @property
    def adjusted_r_squared(self):
        """
        Gets the adjusted_r_squared of this RegressionOutputV1.
        The measure of how close the data is to the regression line, adjusted for the number of input signals and samples

        :return: The adjusted_r_squared of this RegressionOutputV1.
        :rtype: float
        """
        return self._adjusted_r_squared

    @adjusted_r_squared.setter
    def adjusted_r_squared(self, adjusted_r_squared):
        """
        Sets the adjusted_r_squared of this RegressionOutputV1.
        The measure of how close the data is to the regression line, adjusted for the number of input signals and samples

        :param adjusted_r_squared: The adjusted_r_squared of this RegressionOutputV1.
        :type: float
        """
        if adjusted_r_squared is None:
            raise ValueError("Invalid value for `adjusted_r_squared`, must not be `None`")

        self._adjusted_r_squared = adjusted_r_squared

    @property
    def error_sum_squares(self):
        """
        Gets the error_sum_squares of this RegressionOutputV1.
        The standard error for the sum squares

        :return: The error_sum_squares of this RegressionOutputV1.
        :rtype: float
        """
        return self._error_sum_squares

    @error_sum_squares.setter
    def error_sum_squares(self, error_sum_squares):
        """
        Sets the error_sum_squares of this RegressionOutputV1.
        The standard error for the sum squares

        :param error_sum_squares: The error_sum_squares of this RegressionOutputV1.
        :type: float
        """
        if error_sum_squares is None:
            raise ValueError("Invalid value for `error_sum_squares`, must not be `None`")

        self._error_sum_squares = error_sum_squares

    @property
    def intercept(self):
        """
        Gets the intercept of this RegressionOutputV1.
        The constant offset to add. 0 if forceThroughZero was true. This is the intercept for the output signal rather than the individual coefficients.

        :return: The intercept of this RegressionOutputV1.
        :rtype: float
        """
        return self._intercept

    @intercept.setter
    def intercept(self, intercept):
        """
        Sets the intercept of this RegressionOutputV1.
        The constant offset to add. 0 if forceThroughZero was true. This is the intercept for the output signal rather than the individual coefficients.

        :param intercept: The intercept of this RegressionOutputV1.
        :type: float
        """
        if intercept is None:
            raise ValueError("Invalid value for `intercept`, must not be `None`")

        self._intercept = intercept

    @property
    def intercept_standard_error(self):
        """
        Gets the intercept_standard_error of this RegressionOutputV1.
        The standard error for the intercept

        :return: The intercept_standard_error of this RegressionOutputV1.
        :rtype: float
        """
        return self._intercept_standard_error

    @intercept_standard_error.setter
    def intercept_standard_error(self, intercept_standard_error):
        """
        Sets the intercept_standard_error of this RegressionOutputV1.
        The standard error for the intercept

        :param intercept_standard_error: The intercept_standard_error of this RegressionOutputV1.
        :type: float
        """
        if intercept_standard_error is None:
            raise ValueError("Invalid value for `intercept_standard_error`, must not be `None`")

        self._intercept_standard_error = intercept_standard_error

    @property
    def is_uncertain(self):
        """
        Gets the is_uncertain of this RegressionOutputV1.
        True if this regression is uncertain

        :return: The is_uncertain of this RegressionOutputV1.
        :rtype: bool
        """
        return self._is_uncertain

    @is_uncertain.setter
    def is_uncertain(self, is_uncertain):
        """
        Sets the is_uncertain of this RegressionOutputV1.
        True if this regression is uncertain

        :param is_uncertain: The is_uncertain of this RegressionOutputV1.
        :type: bool
        """
        if is_uncertain is None:
            raise ValueError("Invalid value for `is_uncertain`, must not be `None`")

        self._is_uncertain = is_uncertain

    @property
    def r_squared(self):
        """
        Gets the r_squared of this RegressionOutputV1.
        The measure of how close the data is to the regression line

        :return: The r_squared of this RegressionOutputV1.
        :rtype: float
        """
        return self._r_squared

    @r_squared.setter
    def r_squared(self, r_squared):
        """
        Sets the r_squared of this RegressionOutputV1.
        The measure of how close the data is to the regression line

        :param r_squared: The r_squared of this RegressionOutputV1.
        :type: float
        """

        self._r_squared = r_squared

    @property
    def regression_sum_squares(self):
        """
        Gets the regression_sum_squares of this RegressionOutputV1.
        The measure of how well the model matches the target

        :return: The regression_sum_squares of this RegressionOutputV1.
        :rtype: float
        """
        return self._regression_sum_squares

    @regression_sum_squares.setter
    def regression_sum_squares(self, regression_sum_squares):
        """
        Sets the regression_sum_squares of this RegressionOutputV1.
        The measure of how well the model matches the target

        :param regression_sum_squares: The regression_sum_squares of this RegressionOutputV1.
        :type: float
        """
        if regression_sum_squares is None:
            raise ValueError("Invalid value for `regression_sum_squares`, must not be `None`")

        self._regression_sum_squares = regression_sum_squares

    @property
    def suggested_p_value_cutoff(self):
        """
        Gets the suggested_p_value_cutoff of this RegressionOutputV1.
        The value which the regression method suggests for ignoring coefficients

        :return: The suggested_p_value_cutoff of this RegressionOutputV1.
        :rtype: float
        """
        return self._suggested_p_value_cutoff

    @suggested_p_value_cutoff.setter
    def suggested_p_value_cutoff(self, suggested_p_value_cutoff):
        """
        Sets the suggested_p_value_cutoff of this RegressionOutputV1.
        The value which the regression method suggests for ignoring coefficients

        :param suggested_p_value_cutoff: The suggested_p_value_cutoff of this RegressionOutputV1.
        :type: float
        """
        if suggested_p_value_cutoff is None:
            raise ValueError("Invalid value for `suggested_p_value_cutoff`, must not be `None`")

        self._suggested_p_value_cutoff = suggested_p_value_cutoff

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, RegressionOutputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.3.2-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class MonitorValues(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'count': 'int',
        'fifteen_minute_rate': 'float',
        'five_minute_rate': 'float',
        'max': 'float',
        'mean': 'float',
        'min': 'float',
        'one_minute_rate': 'float',
        'p50': 'float',
        'p75': 'float',
        'p95': 'float',
        'p98': 'float',
        'p99': 'float',
        'p999': 'float',
        'stddev': 'float',
        'value': 'float'
    }

    attribute_map = {
        'count': 'count',
        'fifteen_minute_rate': 'fifteenMinuteRate',
        'five_minute_rate': 'fiveMinuteRate',
        'max': 'max',
        'mean': 'mean',
        'min': 'min',
        'one_minute_rate': 'oneMinuteRate',
        'p50': 'p50',
        'p75': 'p75',
        'p95': 'p95',
        'p98': 'p98',
        'p99': 'p99',
        'p999': 'p999',
        'stddev': 'stddev',
        'value': 'value'
    }

    def __init__(self, count=None, fifteen_minute_rate=None, five_minute_rate=None, max=None, mean=None, min=None, one_minute_rate=None, p50=None, p75=None, p95=None, p98=None, p99=None, p999=None, stddev=None, value=None):
        """
        MonitorValues - a model defined in Swagger
        """

        self._count = None
        self._fifteen_minute_rate = None
        self._five_minute_rate = None
        self._max = None
        self._mean = None
        self._min = None
        self._one_minute_rate = None
        self._p50 = None
        self._p75 = None
        self._p95 = None
        self._p98 = None
        self._p99 = None
        self._p999 = None
        self._stddev = None
        self._value = None

        if count is not None:
          self.count = count
        if fifteen_minute_rate is not None:
          self.fifteen_minute_rate = fifteen_minute_rate
        if five_minute_rate is not None:
          self.five_minute_rate = five_minute_rate
        if max is not None:
          self.max = max
        if mean is not None:
          self.mean = mean
        if min is not None:
          self.min = min
        if one_minute_rate is not None:
          self.one_minute_rate = one_minute_rate
        if p50 is not None:
          self.p50 = p50
        if p75 is not None:
          self.p75 = p75
        if p95 is not None:
          self.p95 = p95
        if p98 is not None:
          self.p98 = p98
        if p99 is not None:
          self.p99 = p99
        if p999 is not None:
          self.p999 = p999
        if stddev is not None:
          self.stddev = stddev
        if value is not None:
          self.value = value

    @property
    def count(self):
        """
        Gets the count of this MonitorValues.
        If a timer, the number of timed events

        :return: The count of this MonitorValues.
        :rtype: int
        """
        return self._count

    @count.setter
    def count(self, count):
        """
        Sets the count of this MonitorValues.
        If a timer, the number of timed events

        :param count: The count of this MonitorValues.
        :type: int
        """

        self._count = count

    @property
    def fifteen_minute_rate(self):
        """
        Gets the fifteen_minute_rate of this MonitorValues.
        Rate observed over the last 15 minutes

        :return: The fifteen_minute_rate of this MonitorValues.
        :rtype: float
        """
        return self._fifteen_minute_rate

    @fifteen_minute_rate.setter
    def fifteen_minute_rate(self, fifteen_minute_rate):
        """
        Sets the fifteen_minute_rate of this MonitorValues.
        Rate observed over the last 15 minutes

        :param fifteen_minute_rate: The fifteen_minute_rate of this MonitorValues.
        :type: float
        """

        self._fifteen_minute_rate = fifteen_minute_rate

    @property
    def five_minute_rate(self):
        """
        Gets the five_minute_rate of this MonitorValues.
        Rate observed over the last 5 minutes

        :return: The five_minute_rate of this MonitorValues.
        :rtype: float
        """
        return self._five_minute_rate

    @five_minute_rate.setter
    def five_minute_rate(self, five_minute_rate):
        """
        Sets the five_minute_rate of this MonitorValues.
        Rate observed over the last 5 minutes

        :param five_minute_rate: The five_minute_rate of this MonitorValues.
        :type: float
        """

        self._five_minute_rate = five_minute_rate

    @property
    def max(self):
        """
        Gets the max of this MonitorValues.

        :return: The max of this MonitorValues.
        :rtype: float
        """
        return self._max

    @max.setter
    def max(self, max):
        """
        Sets the max of this MonitorValues.

        :param max: The max of this MonitorValues.
        :type: float
        """

        self._max = max

    @property
    def mean(self):
        """
        Gets the mean of this MonitorValues.

        :return: The mean of this MonitorValues.
        :rtype: float
        """
        return self._mean

    @mean.setter
    def mean(self, mean):
        """
        Sets the mean of this MonitorValues.

        :param mean: The mean of this MonitorValues.
        :type: float
        """

        self._mean = mean

    @property
    def min(self):
        """
        Gets the min of this MonitorValues.

        :return: The min of this MonitorValues.
        :rtype: float
        """
        return self._min

    @min.setter
    def min(self, min):
        """
        Sets the min of this MonitorValues.

        :param min: The min of this MonitorValues.
        :type: float
        """

        self._min = min

    @property
    def one_minute_rate(self):
        """
        Gets the one_minute_rate of this MonitorValues.
        Rate observed over the last 1 minute

        :return: The one_minute_rate of this MonitorValues.
        :rtype: float
        """
        return self._one_minute_rate

    @one_minute_rate.setter
    def one_minute_rate(self, one_minute_rate):
        """
        Sets the one_minute_rate of this MonitorValues.
        Rate observed over the last 1 minute

        :param one_minute_rate: The one_minute_rate of this MonitorValues.
        :type: float
        """

        self._one_minute_rate = one_minute_rate

    @property
    def p50(self):
        """
        Gets the p50 of this MonitorValues.
        50th percentile (also known as median)

        :return: The p50 of this MonitorValues.
        :rtype: float
        """
        return self._p50

    @p50.setter
    def p50(self, p50):
        """
        Sets the p50 of this MonitorValues.
        50th percentile (also known as median)

        :param p50: The p50 of this MonitorValues.
        :type: float
        """

        self._p50 = p50

    @property
    def p75(self):
        """
        Gets the p75 of this MonitorValues.
        75th percentile

        :return: The p75 of this MonitorValues.
        :rtype: float
        """
        return self._p75

    @p75.setter
    def p75(self, p75):
        """
        Sets the p75 of this MonitorValues.
        75th percentile

        :param p75: The p75 of this MonitorValues.
        :type: float
        """

        self._p75 = p75

    @property
    def p95(self):
        """
        Gets the p95 of this MonitorValues.
        95th percentile

        :return: The p95 of this MonitorValues.
        :rtype: float
        """
        return self._p95

    @p95.setter
    def p95(self, p95):
        """
        Sets the p95 of this MonitorValues.
        95th percentile

        :param p95: The p95 of this MonitorValues.
        :type: float
        """

        self._p95 = p95

    @property
    def p98(self):
        """
        Gets the p98 of this MonitorValues.
        98th percentile

        :return: The p98 of this MonitorValues.
        :rtype: float
        """
        return self._p98

    @p98.setter
    def p98(self, p98):
        """
        Sets the p98 of this MonitorValues.
        98th percentile

        :param p98: The p98 of this MonitorValues.
        :type: float
        """

        self._p98 = p98

    @property
    def p99(self):
        """
        Gets the p99 of this MonitorValues.
        99th percentile

        :return: The p99 of this MonitorValues.
        :rtype: float
        """
        return self._p99

    @p99.setter
    def p99(self, p99):
        """
        Sets the p99 of this MonitorValues.
        99th percentile

        :param p99: The p99 of this MonitorValues.
        :type: float
        """

        self._p99 = p99

    @property
    def p999(self):
        """
        Gets the p999 of this MonitorValues.
        99.9th percentile

        :return: The p999 of this MonitorValues.
        :rtype: float
        """
        return self._p999

    @p999.setter
    def p999(self, p999):
        """
        Sets the p999 of this MonitorValues.
        99.9th percentile

        :param p999: The p999 of this MonitorValues.
        :type: float
        """

        self._p999 = p999

    @property
    def stddev(self):
        """
        Gets the stddev of this MonitorValues.
        Standard deviation

        :return: The stddev of this MonitorValues.
        :rtype: float
        """
        return self._stddev

    @stddev.setter
    def stddev(self, stddev):
        """
        Sets the stddev of this MonitorValues.
        Standard deviation

        :param stddev: The stddev of this MonitorValues.
        :type: float
        """

        self._stddev = stddev

    @property
    def value(self):
        """
        Gets the value of this MonitorValues.
        Gauge's current value

        :return: The value of this MonitorValues.
        :rtype: float
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Sets the value of this MonitorValues.
        Gauge's current value

        :param value: The value of this MonitorValues.
        :type: float
        """

        self._value = value

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, MonitorValues):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



from __future__ import annotations

import datetime
import types
from dataclasses import dataclass
from typing import List, Dict, Union, Mapping, Optional, Set

import pandas as pd

from seeq import spy
from seeq.base.seeq_names import SeeqNames
from seeq.sdk import *
from seeq.spy import _common, _login, _metadata, _push, _swap
from seeq.spy._errors import *
from seeq.spy._redaction import safely, request_safely
from seeq.spy._session import Session
from seeq.spy._status import Status

RESERVED_SEARCH_COLUMN_NAMES = ['Path', 'Asset', 'Type', 'Depth', 'Estimated Sample Period', 'Formula Parameters',
                                'Datasource Name']

ALL_PROPERTIES = ['@allProperties']

DEFAULT_PROPERTIES = ['ID', 'Path', 'Asset', 'Name', 'Description', 'Type', 'Value Unit Of Measure', 'Datasource Name',
                      'Archived']

USE_DEFAULT_LIMIT = -1


@dataclass
class SearchContext:
    session: Session
    status: Status

    include_properties: List[str]
    workbook: str
    recursive: bool
    ignore_unindexed_properties: bool
    include_archived: bool
    estimate_sample_period: dict
    include_swap_info: bool
    old_asset_format: bool
    order_by: Optional[Union[str, list]]
    limit: int

    status_index: int
    columns: List[str]
    metadata: List
    ids: Set
    comparison: str
    dupe_count: int
    sample_periods: Dict
    workbook_id: str
    pd_start: Optional[pd.Timestamp]
    pd_end: Optional[pd.Timestamp]
    use_search_items_api_for_this_query: bool
    used_search_items_api_at_least_once: bool
    timer: datetime.datetime

    items_api: ItemsApi
    trees_api: TreesApi
    formulas_api: FormulasApi
    displays_api: DisplaysApi
    display_templates_api: DisplayTemplatesApi

    search_by_id_helper_for_all_queries: SearchByIDHelper
    search_by_id_helper_for_tree_children: SearchByIDHelper

    def __init__(self, session: Session, status: Status):
        self.session = session
        self.status = status

        self.columns = list()
        self.metadata = list()
        self.ids = set()
        self.datasource_ids = dict()
        self.dupe_count = 0
        self.sample_periods = dict()

        self.items_api = ItemsApi(session.client)
        self.trees_api = TreesApi(session.client)
        self.formulas_api = FormulasApi(session.client)
        self.displays_api = DisplaysApi(session.client)
        self.display_templates_api = DisplayTemplatesApi(session.client)


search_by_id_helper_page_size = 1000


class SearchByIDHelper:
    """
    In R62 and later, you can supply a list of filters to the search_items API in the form of ID=={ID}. This means we
    can do far fewer round trips to satisfy a query where the user supplies a DataFrame full of IDs. This class
    issues search_items() calls in pages and caches the results for retrieval by various parts of the search code.
    """

    def __init__(self, context: SearchContext, queries: List[dict]):
        self.context = context
        self.queries = queries
        self.page_size = search_by_id_helper_page_size
        self.current_index = -1
        self.page = None

    def get_by_index(self, index: int) -> ItemSearchPreviewV1:
        if index > self.current_index:
            self._get_next_page_via_search_items()

        return self.page[index]

    def _get_next_page_via_search_items(self):
        filters = ['@includeUnsearchable']
        new_current_index = self.current_index
        while new_current_index < len(self.queries) - 1:
            new_current_index += 1
            if not _common.present(self.queries[new_current_index], 'ID'):
                continue

            filters.append(f"ID=={self.queries[new_current_index]['ID']}")

            if len(filters) > self.page_size:
                break

        if self.context.session.options.wants_compatibility_with(188):
            include_properties = ALL_PROPERTIES
        else:
            include_properties = self.context.include_properties

        search_results = self.context.items_api.search_items(
            filters=filters,
            include_properties=include_properties,
            limit=self.page_size
        )

        items = {item.id: item for item in search_results.items}
        self.page = dict()
        while self.current_index < new_current_index:
            self.current_index += 1
            if not _common.present(self.queries[self.current_index], 'ID'):
                continue

            self.page[self.current_index] = items.get(self.queries[self.current_index]['ID'])


@Status.handle_keyboard_interrupt()
def search(query, *, all_properties=False, include_properties: List[str] = None,
           workbook: Optional[str] = _common.DEFAULT_WORKBOOK_PATH, recursive: bool = True,
           ignore_unindexed_properties: bool = True, include_archived: bool = False,
           include_swappable_assets: bool = False, estimate_sample_period: Optional[dict] = None,
           old_asset_format: bool = None, order_by: Union[str, List[str]] = None,
           limit: Optional[int] = USE_DEFAULT_LIMIT, errors: str = None, quiet: bool = None, status: Status = None,
           session: Optional[Session] = None) -> pd.DataFrame:
    """
    Issues a query to the Seeq Server to retrieve metadata for signals,
    conditions, scalars and assets. This metadata can then be used to retrieve
    samples, capsules for a particular time range via spy.pull().

    Parameters
    ----------
    query : {str, dict, list, pd.DataFrame, pd.Series}
        A mapping of property / match-criteria pairs or a Seeq Workbench URL

        If you supply a dict or list of dicts, then the matching
        operations are "contains" (instead of "equal to").

        If you supply a DataFrame or a Series, then the matching
        operations are "equal to" (instead of "contains").

        If you supply a str, it must be the URL of a Seeq Workbench worksheet.
        The retrieved metadata will be the signals, conditions and scalars
        currently present on the Details Panel.

        'Name' and 'Description' fields support wildcard and regular expression
        (regex) matching with the same syntax as within the Data tab in Seeq
        Workbench.

        The 'Path' field allows you to query within an asset tree, where >>
        separates each level from the next. E.g.: 'Path': 'Example >> Cooling*'
        You can use wildcard and regular expression matching at any level but,
        unlike the Name/Description fields, the match must be a "full match",
        meaning that 'Path': 'Example' will match on a root asset tree node of
        'Example' but not 'Example (AF)'.

        Available options are:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        Name                Name of the item (wildcards/regex supported)
        Path                Asset tree path of the item (should not include the
                            "leaf" asset), using ' >> ' hierarchy delimiters
        Asset               Asset name (i.e., the name of the leaf asset) or ID
        Type                The item type. One of 'Signal', 'Condition',
                            'Scalar', 'Asset', 'Chart', 'Metric', 'Workbook',
                            and 'Worksheet'
        Description         Description of the item (wildcards/regex supported)
        Datasource Name     Name of the datasource
        Datasource ID       The datasource ID, which corresponds to the Id
                            field in the connector configuration
        Datasource Class    The datasource class (e.g. 'OSIsoft PI')
        Data ID             The data ID, whose format is managed by the
                            datasource connector
        Cache Enabled       True to find items where data caching is enabled
        Scoped To           The Seeq ID of a workbook such that results are
                            limited to ONLY items scoped to that workbook.
        =================== ===================================================

    all_properties : bool, default False
        Return all item properties in the result. If you would like to specify
        exactly which properties to return (for better performance/less data),
        use include_properties instead. If both all_properties and
        include_properties are omitted, you will get only the properties that
        come "for free" (no performance penalty) with the query.

    include_properties : list, optional
        A list of extra properties to include in the results. If omitted, the
        default set of properties will be returned. If both all_properties and
        include_properties are omitted, you will get only the properties that
        come "for free" (no performance penalty) with the query.

    workbook : {str, None}, default 'Data Lab >> Data Lab Analysis'
        A path string (with ' >> ' delimiters) or an ID to indicate a workbook
        such that, in addition to globally-scoped items, the workbook's scoped
        items will also be returned in the results.

        If you want all items regardless of scope, use
        workbook=spy.GLOBALS_AND_ALL_WORKBOOKS

        If you want only globally-scoped items, use
        workbook=spy.GLOBALS_ONLY

        If you don't want globally-scoped items in your results, use the
        'Scoped To' field in the 'query' argument instead. (See 'query'
        argument documentation above.)

        The ID for a workbook is visible in the URL of Seeq Workbench, directly
        after the "workbook/" part.

    recursive : bool, default True
        If True, searches that include a Path entry will include items at and
        below the specified location in an asset tree. If False, then only
        items at the specified level will be returned. To get only the root
        assets, supply a Path value of ''.

    ignore_unindexed_properties : bool, default True
        If False, a ValueError will be raised if any properties are supplied
        that cannot be used in the search.

    include_archived : bool, default False
        If True, includes trashed/archived items in the output.

    include_swappable_assets : bool, default False
        Adds a "Swappable Assets" column to the output where each cell is an
        embedded DataFrame that includes the assets that the item refers to and
        can theoretically be swapped for other assets using spy.swap().

    estimate_sample_period : dict, default None
        A dict with the keys 'Start' and 'End'. If provided, an estimated
        sample period for all signals will be included in the output. The
        values for the 'Start' and 'End' keys must be a string that
        pandas.to_datetime() can parse, or a pandas.Timestamp. The start
        and end times are used to bound the calculation of the sample period.
        If the start and end times encompass a time range that is insufficient
        to determine the sample period, a pd.NaT will be returned.
        If the value of 'Start' is set to None, it will default to the value
        of 'End' minus 1 hour. Conversely, if the value of 'End' is set to
        None, it will default to now.

    old_asset_format : bool, default True
        Historically, spy.search() returned rows with a "Type" of "Asset"
        whereby the "Asset" column was the name of the parent asset. This is
        inconsistent with all other aspects of SPy, including
        spy.push(metadata). If you would like Asset rows to instead be
        consistent with the rest of SPy (whereby the "Asset" column is the
        name of the current asset, not the parent), pass in False for this
        argument.

    order_by : {str, list}, default None
        An optional field or list of fields used to sort the search results.
        Fields on which results can be sorted are 'ID', 'Name', and
        'Description'.

    limit : int, default 1000
        A limit on the number of results returned. By default, the limit is
        1000. Specify limit=None to return all results.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedence.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with rows for each item found and columns for each
        property.

        Additionally, the following properties are stored on the "spy"
        attribute of the output DataFrame:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        func                A str value of 'spy.search'
        kwargs              A dict with the values of the input parameters
                            passed to spy.search to get the output DataFrame
        old_asset_format    True if the old Asset format was used (see doc for
                            old_asset_format argument)
        status              A spy.Status object with the status of the
                            spy.search call
        =================== ===================================================

    Examples
    --------
    Search for signals with the name 'Humid' on the asset tree under
    'Example >> Cooling Tower 1', retrieving all properties on the results:

    >>> search_results = spy.search({'Name': 'Humid', 'Path': 'Example >> Cooling Tower 1'}, all_properties=True)

    To access the stored properties:
    >>> search_results.spy.kwargs
    >>> search_results.spy.status

    Search for signals that have a name that starts with 'Area' in the
    datasource 'Example Data' and determine the sample period of each signal
    during the month of January 2018

    >>> search_results = spy.search({
    >>>    'Name': 'Area ?_*',
    >>>    'Datasource Name': 'Example Data'
    >>> }, estimate_sample_period=dict(Start='2018-01-01', End='2018-02-01'))

    Using a pandas.DataFrame as the input:

    >>> my_items = pd.DataFrame(
    >>>     {'Name': ['Area A_Temperature', 'Area B_Compressor Power', 'Optimize' ],
    >>>      'Datasource Name': 'Example Data'})
    >>> spy.search(my_items)

    Using a URL from a Seeq Workbench worksheet:

    >>> my_worksheet_items = spy.search(
    >>> 'https://seeq.com/workbook/17F31703-F0B6-4C8E-B7FD-E20897BD4819/worksheet/CE6A0B92-EE00-45FC-9EB3-D162632DBB48')

    """
    input_args = _common.validate_argument_types([
        (query, 'query', (str, dict, list, pd.DataFrame, pd.Series)),
        (all_properties, 'all_properties', bool),
        (include_properties, 'include_properties', list),
        (workbook, 'workbook', str),
        (recursive, 'recursive', bool),
        (ignore_unindexed_properties, 'ignore_unindexed_properties', bool),
        (include_archived, 'include_archived', bool),
        (estimate_sample_period, 'estimate_sample_period', dict),
        (include_swappable_assets, 'include_swappable_assets', bool),
        (old_asset_format, 'old_asset_format', bool),
        (order_by, 'order_by', (str, list)),
        (limit, 'limit', int),
        (quiet, 'quiet', bool),
        (errors, 'errors', str),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet, errors)
    session = Session.validate(session)
    _login.validate_login(session, status)

    if all_properties:
        if include_properties is not None:
            status.warn(f'If you specify all_properties=True, the include_properties argument is ignored.')
        include_properties = ALL_PROPERTIES.copy()

    context = SearchContext(session, status)
    context.workbook = workbook
    context.recursive = recursive
    context.ignore_unindexed_properties = ignore_unindexed_properties
    context.include_archived = include_archived
    context.estimate_sample_period = estimate_sample_period
    context.include_swap_info = include_swappable_assets
    context.old_asset_format = old_asset_format
    context.order_by = order_by

    if limit != USE_DEFAULT_LIMIT:
        context.limit = limit
    elif session.options.wants_compatibility_with(188):
        # In v188 and earlier, there was no limiting capability.
        context.limit = None
    else:
        # In v189 and later, we apply a default limit of 1000 to reduce the likelihood of a user accidentally running
        # a search that returns a huge number of results.
        context.limit = 1000

    context.columns = ['ID', 'Type']
    if include_properties is not None and include_properties != ALL_PROPERTIES:
        context.columns.extend(include_properties)

    context.include_properties = include_properties if include_properties is not None else list()

    # The SPy user knows the formula parameters property as "Formula Parameters", but the backend knows it as
    # "FormulaParameters". So just convert it here to keep things consistent for the SPy user.
    context.include_properties = [p.replace('Formula Parameters', SeeqNames.Properties.formula_parameters)
                                  for p in context.include_properties]

    if context.order_by:
        context.order_by = _validate_order_by(context.order_by)

    if context.estimate_sample_period is not None:
        if context.estimate_sample_period.keys() != {'Start', 'End'}:  # strict comparison, allowing only these two keys
            raise SPyValueError(f"estimate_sample_period must have 'Start' and 'End' keys but got "
                                f"{context.estimate_sample_period.keys()}")
        context.pd_start, context.pd_end = _login.validate_start_and_end(context.session,
                                                                         context.estimate_sample_period['Start'],
                                                                         context.estimate_sample_period['End'])

    if not context.recursive and 'Path' not in query:
        raise SPyValueError("'Path' must be included in query when recursive=False")

    context.old_asset_format = old_asset_format
    if context.old_asset_format is None:
        # In the future, we may wish to change this default to False, in which case we should use
        # spy.options.compatibility and keep it True for users expecting older behavior.
        context.old_asset_format = True

    queries: List[Union[Dict, Mapping]]
    if isinstance(query, pd.DataFrame):
        queries = query.to_dict(orient='records')
        context.comparison = '=='
    elif isinstance(query, pd.Series):
        queries = [query.to_dict()]
        context.comparison = '=='
    elif isinstance(query, list):
        queries = query
        context.comparison = '~='
    elif isinstance(query, str):
        worksheet = spy.utils.get_analysis_worksheet_from_url(query, context.include_archived,
                                                              quiet=context.status.quiet)
        queries = worksheet.display_items.to_dict(orient='records')
        context.comparison = '=='
    else:
        queries = [query]
        context.comparison = '~='

    # Handle the case where the user provides a lower-case ID (CRAB-38168)
    for query in queries:
        if _common.present(query, 'ID'):
            query['ID'] = query['ID'].upper()

    context.status.df = pd.DataFrame(queries)
    context.status.df['Time'] = 0
    context.status.df['Count'] = 0
    context.status.df['Pages'] = 0
    context.status.df['Result'] = 'Queued'
    context.status.update('Initializing', Status.RUNNING)

    context.workbook_id = None
    if context.workbook:
        if _common.is_guid(context.workbook):
            context.workbook_id = _common.sanitize_guid(context.workbook)
        else:
            search_query, _ = _push.create_analysis_search_query(context.workbook)
            search_df = spy.workbooks.search(search_query,
                                             status=context.status.create_inner('Find Workbook', quiet=True),
                                             session=context.session)
            context.workbook_id = search_df.iloc[0]['ID'] if len(search_df) > 0 else None
            if context.workbook == _common.DEFAULT_WORKBOOK_PATH and context.workbook_id is None:
                context.workbook_id = _common.GLOBALS_ONLY

    context.search_by_id_helper_for_all_queries = SearchByIDHelper(context, queries)

    context.used_search_items_api_at_least_once = False

    try:
        for status_index in range(len(queries)):
            context.status_index = status_index
            _process_query(context, queries[status_index])
    except StopIteration:
        status.warn(f'Search results limited to {context.limit}. Supply increased "limit" parameter to return more.')

    if context.dupe_count > 0:
        if context.used_search_items_api_at_least_once:
            arg_to_use = context.order_by if context.order_by is not None else list()
            if 'ID' in arg_to_use:
                context.status.warn(
                    f'{context.dupe_count} duplicates removed from returned DataFrame. If you used a list of '
                    f'searches, those searches may have had overlap.')
            else:
                arg_to_use.append('ID')
                context.status.warn(
                    f'{context.dupe_count} duplicates removed from returned DataFrame. Use order_by={arg_to_use} in '
                    f'your spy.search to ensure results are not missing any items.')
        else:
            context.status.warn(f'{context.dupe_count} duplicates removed from returned DataFrame.')

    context.status.update('Query successful', Status.SUCCESS)

    # We want the default properties to always appear first and in a certain order, so we make a final_column_set
    # list, add in the additional properties, and then filter any absent properties from the final list
    final_column_set = DEFAULT_PROPERTIES.copy()
    final_column_set += [c for c in context.columns if c not in final_column_set]
    final_column_set = [c for c in final_column_set if c in context.columns]
    output_df = pd.DataFrame(data=context.metadata, columns=final_column_set)

    if context.order_by and not output_df.empty:
        output_df.sort_values(context.order_by, ignore_index=True, inplace=True)

    if old_asset_format is None:
        type_column = output_df['Type']
        if len(type_column.loc[type_column == 'Asset']) > 0:
            context.status.warn(
                'This search result includes Assets. Consider passing in "old_asset_format=False" so that the '
                '"Path" and "Asset" columns are populated in a way that is consistent with all other aspects '
                'of SPy. The default behavior without this argument emulates the (incorrect) way SPy would '
                'return such results historically.')

    output_df_properties = types.SimpleNamespace(
        func='spy.search',
        kwargs=input_args,
        old_asset_format=context.old_asset_format,
        status=context.status)

    _common.put_properties_on_df(output_df, output_df_properties)

    return output_df


def _process_query(context: SearchContext, current_query):
    """
    Called for each query "row" in the list of queries
    """
    context.timer = _common.timer_start()

    if _common.present(current_query, 'ID'):
        context.search_by_id_on_this_query = True

        # If ID is specified, short-circuit everything and just get the item directly.
        current_id = current_query['ID']
        prop_dict = _create_prop_dict(context, current_id)
        _add_to_metadata(context, prop_dict)

        context.status.df.at[context.status_index, 'Time'] = _common.timer_elapsed(context.timer)
        context.status.df.at[context.status_index, 'Count'] = 1
        context.status.df.at[context.status_index, 'Result'] = 'Success'
        return

    context.search_by_id_on_this_query = False

    # If the user wants a recursive search or there's no 'Path' in the query, then use the ItemsApi.search_items API
    context.use_search_items_api_for_this_query = context.recursive or not _common.present(current_query, 'Path')
    context.used_search_items_api_at_least_once = (context.used_search_items_api_at_least_once or
                                                   context.use_search_items_api_for_this_query)

    if not context.use_search_items_api_for_this_query and context.include_archived:
        # As you can see in the code below, the TreesApi.get_tree() API doesn't have the ability to request
        # archived items
        raise SPyValueError('include_archived=True can only be used with recursive searches or searches that do '
                            'not involve a Path parameter')

    allowed_properties = ['Type', 'Name', 'Description', 'Path', 'Asset', 'Datasource Class', 'Datasource ID',
                          'Datasource Name', 'Data ID', 'Cache Enabled', 'Scoped To']

    disallowed_properties = list()
    for key, value in current_query.items():
        if key not in allowed_properties:
            disallowed_properties.append(key)

    for key in disallowed_properties:
        del current_query[key]

    allowed_properties_str = '", "'.join(allowed_properties)
    if len(disallowed_properties) > 0:
        disallowed_properties_str = '", "'.join(disallowed_properties)
        message = f'The following properties are not indexed' \
                  f'{" and will be ignored" if context.ignore_unindexed_properties else ""}:\n' \
                  f'"{disallowed_properties_str}"\n' \
                  f'Use any of the following searchable properties and then filter further using DataFrame ' \
                  f'operations:\n"{allowed_properties_str}"'

        if not context.ignore_unindexed_properties:
            raise SPyValueError(message)
        else:
            context.status.warn(message)

    if len(current_query) == 0:
        raise SPyValueError('No recognized properties present in "query" argument. You must provide a dict or a '
                            'DataFrame with keys or columns with at least one of the following properties: \n' +
                            allowed_properties_str)

    item_types = list()
    clauses: Dict = dict()

    if _common.present(current_query, 'Type'):
        item_type_specs = list()
        if isinstance(current_query['Type'], list):
            item_type_specs.extend(current_query['Type'])
        else:
            item_type_specs.append(current_query['Type'])

        valid_types = ['StoredSignal', 'CalculatedSignal',
                       'StoredCondition', 'CalculatedCondition',
                       'LiteralScalar', 'CalculatedScalar',
                       'Datasource',
                       'ThresholdMetric', 'Chart', 'Asset',
                       'Workbook', 'Worksheet',
                       'Display', 'DisplayTemplate']

        for item_type_spec in item_type_specs:
            if item_type_spec == 'Signal':
                item_types.extend(['StoredSignal', 'CalculatedSignal'])
            elif item_type_spec == 'Condition':
                item_types.extend(['StoredCondition', 'CalculatedCondition'])
            elif item_type_spec == 'Scalar':
                item_types.extend(['LiteralScalar', 'CalculatedScalar'])
            elif item_type_spec == 'Datasource':
                item_types.extend(['Datasource'])
            elif item_type_spec == 'Metric':
                item_types.extend(['ThresholdMetric'])
            elif item_type_spec not in valid_types:
                raise SPyValueError(f'Type field value not recognized: {item_type_spec}\n'
                                    f'Valid types: {", ".join(valid_types)}')
            else:
                item_types.append(item_type_spec)

        del current_query['Type']

    for prop_name in ['Name', 'Description', 'Datasource Class', 'Datasource ID', 'Data ID']:
        if prop_name in current_query and not pd.isna(current_query[prop_name]):
            clauses[prop_name] = (context.comparison, current_query[prop_name])

    if _common.present(current_query, 'Datasource Name'):
        datasource_name = _common.get(current_query, 'Datasource Name')
        if datasource_name in context.datasource_ids:
            clauses['Datasource ID'], clauses['Datasource Class'] = context.datasource_ids[datasource_name]
        else:
            filters = ['Name == %s' % datasource_name]
            if _common.present(current_query, 'Datasource ID'):
                filters.append('Datasource ID == %s' % _common.get(current_query, 'Datasource ID'))
            if _common.present(current_query, 'Datasource Class'):
                filters.append('Datasource Class == %s' % _common.get(current_query, 'Datasource Class'))

            filter_list = [' && '.join(filters)]
            if context.include_archived:
                filter_list.append('@includeUnsearchable')

            datasource_results = context.items_api.search_items(
                filters=filter_list, types=['Datasource'], limit=100000)

            if len(datasource_results.items) > 1:
                raise SPyRuntimeError('Multiple datasources found that match "%s"' % datasource_name)
            elif len(datasource_results.items) == 0:
                raise SPyRuntimeError('No datasource found that matches "%s"' % datasource_name)

            datasource = datasource_results.items[0]

            @request_safely(action_description=f'get datasource details for "{datasource_name}" {datasource.id}',
                            status=context.status)
            def request_datasource_and_set_clause():
                property_output = context.items_api.get_property(id=datasource.id, property_name='Datasource Class')
                clauses['Datasource Class'] = ('==', property_output.value)
                property_output = context.items_api.get_property(id=datasource.id, property_name='Datasource ID')
                clauses['Datasource ID'] = ('==', property_output.value)
                context.datasource_ids[datasource_name] = (clauses['Datasource ID'], clauses['Datasource Class'])

            request_datasource_and_set_clause()

        del current_query['Datasource Name']

    filters = list()
    if len(clauses.items()) > 0:
        filters.append(' && '.join([p + c + v for p, (c, v) in clauses.items()]))

    if context.include_archived:
        filters.append('@includeUnsearchable')

    kwargs = {
        'filters': filters,
        'types': item_types
    }

    if context.limit is None:
        kwargs['limit'] = context.session.options.search_page_size
    else:
        # If the user supplied a limit argument, then use it (if it's smaller than the page size)
        kwargs['limit'] = min(context.session.options.search_page_size, context.limit)

    if context.include_properties is not None and _login.is_sdk_module_version_at_least(62):
        kwargs['include_properties'] = context.include_properties

    if context.workbook:
        if context.workbook_id:
            kwargs['scope'] = context.workbook_id
        elif context.workbook != _common.DEFAULT_WORKBOOK_PATH:
            raise SPyRuntimeError(f'Workbook "{context.workbook}" not found, or is not accessible by you')

    if _common.present(current_query, 'Scoped To'):
        kwargs['scope'] = current_query['Scoped To']
        kwargs['filters'].append('@excludeGloballyScoped')

    if _common.present(current_query, 'Asset'):
        if _common.is_guid(_common.get(current_query, 'Asset')):
            kwargs['asset'] = _common.get(current_query, 'Asset')
        elif not _common.present(current_query, 'Path'):
            raise SPyValueError('"Path" query parameter must be present when "Asset" name parameter present')

    path_to_query = None
    if _common.present(current_query, 'Path'):
        path_to_query = current_query['Path']
        if _common.present(current_query, 'Asset'):
            path_to_query = path_to_query + ' >> ' + current_query['Asset']

    if not _common.present(current_query, 'Path'):
        # If there's no 'Path' property in the query, we can immediately proceed to the results gathering stage.
        _gather_results(context, current_query, kwargs)
    else:
        # If there is a 'Path' property in the query, then first we have to drill down through the tree to the
        # appropriate depth so we can find the asset ID to use for the results gathering stage.

        if len(path_to_query) == 0:
            _gather_results(context, current_query, kwargs, list())
        else:
            _process_query_path_string(context, current_query, kwargs, path_to_query, list(), clauses)

    context.status.df.at[context.status_index, 'Result'] = 'Success'


def _gather_results(context: SearchContext, current_query, kwargs, actual_path_list=None):
    """
    Once we're at the appropriate depth in the tree, we can gather the results. If we're not doing a tree-based
    query, then we just search at the top level.
    """
    if context.use_search_items_api_for_this_query:
        _iterate_over_output(context, _do_search, 'items', _gather_results_via_item_search, current_query,
                             kwargs, actual_path_list)
    else:
        _iterate_over_output(context, _do_search, 'children', _gather_results_via_get_tree, current_query,
                             kwargs, actual_path_list)


# noinspection PyUnusedLocal
def _iterate_over_output(context: SearchContext, _output_func, _collection_name, _action_func, current_query,
                         kwargs, _actual_path_list):
    """
    Handles pagination for the various ways that we need to churn through results.
    """
    offset = 0
    while True:
        output = _output_func(context, kwargs, offset)

        collection = getattr(output, _collection_name)

        context.status.df.at[context.status_index, 'Time'] = _common.timer_elapsed(context.timer)
        context.status.df.at[context.status_index, 'Count'] = offset + len(collection)
        context.status.df.at[context.status_index, 'Pages'] += 1
        context.status.df.at[context.status_index, 'Result'] = 'Querying'
        context.status.update('Querying Seeq Server for items', Status.RUNNING)

        for item in collection:
            _action_func(context, current_query, _actual_path_list, item)

        if len(collection) != output.limit:
            break

        offset += output.limit


def _do_search(context: SearchContext, kwargs: dict, offset):
    """
    Executes an actual search. Sometimes this uses the SearchItems API, sometimes it uses the GetTree API.
    """
    kwargs['offset'] = offset
    if 'scope' in kwargs and isinstance(kwargs['scope'], str):
        kwargs['scope'] = [kwargs['scope']]
    if context.use_search_items_api_for_this_query:
        if context.order_by:
            kwargs['order_by'] = context.order_by
        return context.items_api.search_items(**kwargs)

    kwargs2 = {
        'offset': kwargs['offset'],
        'limit': kwargs['limit'],
        'scope': _common.get(kwargs, 'scope'),
        'exclude_globally_scoped': ('@excludeGloballyScoped' in kwargs['filters'])
    }

    if 'asset' in kwargs:
        kwargs2['id'] = kwargs['asset']
        tree_output = context.trees_api.get_tree(**kwargs2)
    else:
        tree_output = context.trees_api.get_tree_root_nodes(**kwargs2)
    if len(kwargs['types']) > 0:
        tree_output.children = [x for x in tree_output.children if x.type in kwargs['types']]

    # This will speed up retrieval of properties in R62 and later
    for i in range(len(tree_output.children)):
        # Plant an index on the child object so we can find it later
        setattr(tree_output.children[i], 'child_index', i)

    # Create a SearchByIDHelper that will be called (in R62 and later) if we need to retrieve extra properties for the
    # child. This case only happens when performing a tree-based query with recursive=False.
    context.search_by_id_helper_for_tree_children = SearchByIDHelper(context, [{'ID': child.id} for child in
                                                                               tree_output.children])

    return tree_output


# noinspection PyUnusedLocal
def _gather_results_via_item_search(context: SearchContext, current_query, actual_path_list,
                                    result: ItemSearchPreviewV1):
    item_search_preview: ItemSearchPreviewV1 = result
    prop_dict = _create_prop_dict(context, item_search_preview)
    _add_to_metadata(context, prop_dict)


def _gather_results_via_get_tree(context: SearchContext, current_query, actual_path_list, result):
    tree_item_output: TreeItemOutputV1 = result

    for prop, _attr in [('Name', 'name'), ('Description', 'description')]:
        if prop not in current_query:
            continue

        if not _common.does_query_fragment_match(current_query[prop],
                                                 getattr(tree_item_output, _attr),
                                                 contains=(context.comparison == '~=')):
            return

    prop_dict = _create_prop_dict(context, tree_item_output)

    _add_ancestors_to_prop_dict(
        context, tree_item_output.type, tree_item_output.name, actual_path_list, prop_dict)

    _add_to_metadata(context, prop_dict)


def _create_prop_dict(context: SearchContext, item: Union[str, TreeItemOutputV1, ItemSearchPreviewV1]):
    """
    Each individual search result row is represented by a dictionary. This function creates that dictionary,
    and handles the various permutations of types that we might encounter.
    """

    def _add_error_message_and_warn(msg):
        _add_to_dict(context, prop_dict, 'Pull Result', msg)
        context.status.warn(msg)

    item_id = item if isinstance(item, str) else item.id

    prop_dict = dict()

    @request_safely(action_description=f'get included item properties for {item_id}',
                    status=context.status,
                    on_error=_add_error_message_and_warn)
    def _request_item_properties():
        output_object = _add_properties_smartly(context, item, prop_dict)

        if output_object.type in ['CalculatedSignal', 'CalculatedCondition', 'CalculatedScalar', 'LiteralScalar']:
            formula_parameters = None
            if _login.is_sdk_module_version_at_least(62):
                formula_parameters = output_object.parameters
            elif (_has_requested_property(context, 'Formula') or
                  _has_requested_property(context, 'FormulaParameters')):
                # Formula isn't always included in the output (in this R61-and-earlier-case), and Formula Parameters
                # is never included. So we have to make a separate request.
                formula_output = context.formulas_api.get_item(id=output_object.id)
                if _has_requested_property(context, 'Formula'):
                    _add_to_dict(context, prop_dict, 'Formula', formula_output.formula)
                formula_parameters = formula_output.parameters

            if formula_parameters is not None and _has_requested_property(context, 'FormulaParameters'):
                _add_to_dict(context, prop_dict, 'Formula Parameters', [
                    '%s=%s' % (_p.name, _p.item.id if _p.item else _p.formula) for _p in formula_parameters
                ])

        elif output_object.type == 'ThresholdMetric':
            formula_parameters = _metadata.formula_parameters_dict_from_threshold_metric(
                context.session, output_object.id)
            for key, value in formula_parameters.items():
                _add_to_dict(context, prop_dict, key, value)

        elif output_object.type == 'Display':
            display_output = context.displays_api.get_display(id=output_object.id)
            _add_to_dict(context, prop_dict, 'Template ID', display_output.template.id)
            if display_output.swap is not None:
                _add_to_dict(context, prop_dict, 'Swap Out Asset ID', display_output.swap.swap_out)
                _add_to_dict(context, prop_dict, 'Swap In Asset ID', display_output.swap.swap_in)
            else:
                _add_to_dict(context, prop_dict, 'Swap Out Asset ID', display_output.template.swap_source_asset_id)

        elif output_object.type == 'DisplayTemplate':
            display_template_output = context.display_templates_api.get_display_template(id=output_object.id)
            _add_to_dict(context, prop_dict, 'Source Workstep ID', display_template_output.source_workstep_id)

    _request_item_properties()
    return prop_dict


def _validate_order_by(order_by: Union[str, List[str]]) -> List[str]:
    """
    Validate and process order_by arg of spy.search
    """
    # convert string order_by to a list
    if isinstance(order_by, str):
        order_by = [order_by]

    # validate order_by
    order_fields = ['ID', 'Name', 'Description']
    invalid_fields = [x for x in order_by if x not in order_fields]
    if len(invalid_fields) > 0:
        raise SPyValueError(
            f"Invalid order_by fields: {invalid_fields}. Search results can only be ordered on "
            f"{order_fields} fields.")

    return order_by


def _add_properties_smartly(context: SearchContext,
                            item: Union[str, TreeItemOutputV1, ItemSearchPreviewV1],
                            prop_dict: dict):
    # This function does a bunch of heavy lifting to handle the differences in how properties are returned
    # in different versions of the SDK.
    #
    # R62 and later:   Extra properties are returned in ItemSearchPreviewV1 objects
    # R61 and earlier: Extra properties must be fetched in a separate request

    # The following two variables will be filled in if we find that we already had them or needed them to fulfill the
    # user's request.
    item_search_preview: Optional[ItemSearchPreviewV1] = None
    item_output: Optional[ItemOutputV1] = None

    # This is the variable that we use to grab all the default/common properties
    item_object: Union[ItemSearchPreviewV1, ItemOutputV1] = item

    if isinstance(item, str):
        if _login.is_sdk_module_version_at_least(62):
            item_search_preview = context.search_by_id_helper_for_all_queries.get_by_index(context.status_index)
            item_object = item_search_preview
        else:
            # This does not need to be wrapped in request_safely because it is only called from a safe place in
            # _create_prop_dict()
            item_output = context.items_api.get_item_and_all_properties(id=item)
            item_object = item_output
    elif isinstance(item, ItemSearchPreviewV1):
        item_search_preview = item

    _add_if_available(context, prop_dict, 'ID', item_object, 'id')
    _add_if_available(context, prop_dict, 'Name', item_object, 'name')
    _add_if_available(context, prop_dict, 'Description', item_object, 'description')
    _add_if_available(context, prop_dict, 'Type', item_object, 'type')
    _add_if_available(context, prop_dict, 'Archived', item_object, 'is_archived')
    _add_if_available(context, prop_dict, 'Scoped To', item_object, 'scoped_to')

    if hasattr(item_object, 'value_unit_of_measure'):
        if item_object.value_unit_of_measure is None and hasattr(item_object, 'source_value_unit_of_measure'):
            uom = item_object.source_value_unit_of_measure
        else:
            uom = item_object.value_unit_of_measure

        _add_to_dict(context, prop_dict, 'Value Unit Of Measure', uom)
    elif item_output is not None:
        value_unit_of_measure = [p.value for p in item_output.properties if p.name == 'Value Unit Of Measure']
        source_value_unit_of_measure = [p.value for p in item_output.properties
                                        if p.name == 'Source Value Unit Of Measure']
        if len(value_unit_of_measure) == 0 and len(source_value_unit_of_measure) == 1:
            uom = source_value_unit_of_measure[0]
        elif len(value_unit_of_measure) == 1:
            uom = value_unit_of_measure[0]
        else:
            uom = None

        _add_to_dict(context, prop_dict, 'Value Unit Of Measure', uom)

    if _has_requested_property(context, 'Source Value Unit Of Measure'):
        _add_if_available(context, prop_dict, 'Source Value Unit Of Measure', item_object,
                          'source_value_unit_of_measure')

    if hasattr(item_object, 'datasource'):
        datasource_item_preview: ItemPreviewV1 = item_object.datasource
        _add_to_dict(context, prop_dict, 'Datasource Name',
                     datasource_item_preview.name if datasource_item_preview else None)

    if hasattr(item_object, 'ancestors'):
        _add_ancestors_to_prop_dict_from_item_output(context, item_object, prop_dict)

    if context.include_swap_info or (
            not hasattr(item_object, 'ancestors') and
            (_has_requested_property(context, 'Path') or _has_requested_property(context, 'Asset'))
    ):
        _retrieve_ancestors_and_swap_info(context, item_object.id, prop_dict)

    if context.estimate_sample_period is not None:
        _estimate_sample_period(context, item_object)
        _add_to_dict(context, prop_dict, 'Estimated Sample Period', context.sample_periods[item_object.id])

    # Check if we need any more properties. If we don't, we can avoid more API calls.
    need_more_properties = False
    if isinstance(item, str) and context.session.options.wants_compatibility_with(188):
        need_more_properties = True
    else:
        for prop in context.include_properties:
            if prop not in prop_dict:
                need_more_properties = True
                break

    if not need_more_properties:
        return item_object

    if _login.is_sdk_module_version_at_least(62):
        if item_search_preview is None and isinstance(item, TreeItemOutputV1):
            item_search_preview = context.search_by_id_helper_for_tree_children.get_by_index(
                getattr(item, 'child_index'))

        for prop in item_search_preview.included_properties.values():  # type: ScalarPropertyV1
            _add_included_property_to_dict(context, prop_dict, prop.name, prop.value)
    else:
        if item_output is None:
            # This does not need to be wrapped in request_safely because it is only called from a safe place in
            # _create_prop_dict()
            item_output = context.items_api.get_item_and_all_properties(id=item_object.id)

        for prop in item_output.properties:  # type: PropertyOutputV1
            _add_included_property_to_dict(context, prop_dict, prop.name, prop.value)

    return item_object


def _add_included_property_to_dict(context: SearchContext, prop_dict: dict, prop_name: str, prop_value: object):
    if prop_name in RESERVED_SEARCH_COLUMN_NAMES:
        return

    if not context.session.options.wants_compatibility_with(188) and not _has_requested_property(context, prop_name):
        return

    _add_to_dict(context, prop_dict, prop_name, prop_value)


def _add_if_available(context: SearchContext, prop_dict: dict, key: str, obj: object, attr: str):
    if hasattr(obj, attr):
        _add_to_dict(context, prop_dict, key, getattr(obj, attr))


def _add_to_dict(context: SearchContext, prop_dict: dict, key: str, val: object):
    if context.session.options.wants_compatibility_with(188):
        # In v188 and earlier, all scalar properties on items were returned as strings from
        # items_api.get_item_and_all_properties(). So for scripts that are expecting that, cast it to a string.
        val = str(val)

    elif key.endswith('Unit Of Measure') and val == '':
        # This property is returned as None in ItemSearchPreviewV1.value_unit_of_measure, but as an empty string in
        # the ItemSearchPreviewV1.included_properties. Normalize to None in v189 and later.
        val = None

    if key in ['Archived', 'Cache Enabled', 'Enabled', 'Unsearchable'] and isinstance(val, str):
        val = val.lower() == 'true'

    prop_dict[key] = _common.none_to_nan(val)

    # We want the columns to appear in a certain order (the order we added them in) for readability
    if key not in context.columns:
        context.columns.append(key)


def _add_ancestors_to_prop_dict_from_item_output(context: SearchContext, item_output, prop_dict: dict):
    _ancestors = [a.name for a in item_output.ancestors]
    _add_ancestors_to_prop_dict(context, item_output.type, item_output.name, _ancestors, prop_dict)


def _add_ancestors_to_prop_dict(context: SearchContext, item_type: str, name: str, ancestors: list, prop_dict: dict):
    _common.add_ancestors_to_definition(item_type, name, ancestors, prop_dict, context.old_asset_format)
    for key in ['Path', 'Asset']:
        if key in prop_dict and key not in context.columns:
            context.columns.append(key)


def _add_to_metadata(context: SearchContext, prop_dict: dict):
    if prop_dict['ID'] not in context.ids:
        if len(context.metadata) == context.limit:
            raise StopIteration()

        context.metadata.append(prop_dict)
        context.ids.add(prop_dict['ID'])
    else:
        context.dupe_count += 1


def _estimate_sample_period(context: SearchContext, item: Union[TreeItemOutputV1, ItemSearchPreviewV1]):
    if 'Signal' not in item.type:
        context.sample_periods[item.id] = pd.NaT
        return

    if item.id in context.sample_periods:
        # Already done
        return

    sampling_formula = "$signal.estimateSamplePeriod(" \
                       f"capsule('{context.pd_start.isoformat()}','{context.pd_end.isoformat()}'))"

    formula_run_output = safely(
        lambda: context.formulas_api.run_formula(formula=sampling_formula, parameters=[f"signal={item.id}"]),
        action_description='estimate sample period',
        status=context.status,
        additional_errors=[400])

    if formula_run_output is not None and formula_run_output.scalar.value is not None:
        context.sample_periods[item.id] = pd.to_timedelta(
            formula_run_output.scalar.value, unit=formula_run_output.scalar.uom)
    else:
        context.status.warn(
            f'Could not determine the sample period for signal "{item.name}" {item.id} within the '
            f'time period {context.pd_start.isoformat()} to {context.pd_end.isoformat()}. '
            f'There might not be enough data in the specified time range. Modify the time period with the '
            f'`estimate_start` and `estimate_end` arguments.'
        )
        context.sample_periods[item.id] = pd.NaT


def _has_requested_property(context: SearchContext, prop_name: str) -> bool:
    return prop_name in context.include_properties or context.include_properties == ALL_PROPERTIES


def _retrieve_ancestors_and_swap_info(context: SearchContext, item_id: str, prop_dict: dict):
    """
    This function tries to economically retrieve ancestors and swap information if necessary.
    """
    item_output = None
    if context.include_swap_info:
        item_dependency_output: ItemDependencyOutputV1 = safely(
            lambda: context.items_api.get_formula_dependencies(id=item_id),
            action_description=f'get dependencies for {item_id}',
            status=context.status)
        if item_dependency_output is not None:
            item_output = item_dependency_output

        dependencies_with_relevant_assets = _swap.get_swappable_assets(item_dependency_output)

        def _swappable_asset_dict(d):
            leaf_asset = d.ancestors[-1]
            return {
                'ID': leaf_asset.id,
                'Type': leaf_asset.type,
                'Path': _common.path_list_to_string([a.name for a in d.ancestors[0:-1]]),
                'Asset': leaf_asset.name
            }

        swappable_assets = pd.DataFrame([_swappable_asset_dict(d) for d in dependencies_with_relevant_assets],
                                        columns=['ID', 'Type', 'Path', 'Asset'])

        _add_to_dict(context, prop_dict, 'Swappable Assets', swappable_assets)
    else:
        asset_tree_output: AssetTreeOutputV1 = safely(lambda: context.trees_api.get_tree(id=item_id),
                                                      action_description=f'get asset tree ancestors for {item_id}',
                                                      status=context.status)
        if asset_tree_output is not None:
            item_output = asset_tree_output.item

    if item_output is not None:
        _add_ancestors_to_prop_dict_from_item_output(context, item_output, prop_dict)

    return prop_dict


def _process_query_path_string(context: SearchContext, current_query, kwargs, remaining_query_path_string,
                               actual_path_list, clauses: dict, asset_id=None):
    """
    Walks down an asset path and processes each level, including wildcards and regexes.
    """
    query_path_list = _common.path_string_to_list(remaining_query_path_string)

    query_path_part = query_path_list[0]

    tree_kwargs = dict()
    tree_kwargs['limit'] = kwargs['limit']
    tree_kwargs['offset'] = 0

    if 'scope' in kwargs and isinstance(kwargs['scope'], str):
        tree_kwargs['scope'] = [kwargs['scope']]

    while True:
        tree_output: AssetTreeOutputV1
        if not asset_id:
            tree_output = context.trees_api.get_tree_root_nodes(**tree_kwargs)
        else:
            tree_kwargs['id'] = asset_id
            tree_output = context.trees_api.get_tree(**tree_kwargs)

        for child in tree_output.children:  # type: TreeItemOutputV1
            if not _common.does_query_fragment_match(query_path_part, child.name, contains=False):
                continue

            if not asset_id:
                @request_safely(action_description=f'check if "{child.name}" {child.id} has '
                                                   f'datasource matching request',
                                status=context.status,
                                default_value=True)
                def _is_item_in_filter_datasources():
                    child_item_output = context.items_api.get_item_and_all_properties(id=child.id)
                    for prop in ['Datasource Class', 'Datasource ID']:
                        if prop in clauses:
                            _, val = clauses[prop]
                            p_list = [_p.value for _p in child_item_output.properties if
                                      _p.name == prop]
                            if len(p_list) == 0 or p_list[0] != val:
                                return False
                    return True

                # We only filter out datasource at the top level, in case the tree is mixed
                if not _is_item_in_filter_datasources():
                    continue

            actual_path_list_for_child = actual_path_list.copy()
            actual_path_list_for_child.append(child.name)
            if len(query_path_list) == 1:
                kwargs['asset'] = child.id
                _gather_results(context, current_query, kwargs, actual_path_list=actual_path_list_for_child)
            else:
                _process_query_path_string(context,
                                           current_query,
                                           kwargs,
                                           _common.path_list_to_string(query_path_list[1:]),
                                           actual_path_list_for_child,
                                           clauses,
                                           child.id)

        if len(tree_output.children) < tree_kwargs['limit']:
            break

        tree_kwargs['offset'] += tree_kwargs['limit']



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 59.3.2-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class TreeItemOutputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cached': 'bool',
        'children': 'list[TreeItemOutputV1]',
        'created_by': 'str',
        'datasource': 'DatasourcePreviewV1',
        'description': 'str',
        'effective_permissions': 'PermissionsV1',
        'fingerprint': 'str',
        'has_children': 'bool',
        'id': 'str',
        'is_archived': 'bool',
        'is_redacted': 'bool',
        'manually_added': 'str',
        'name': 'str',
        'parameters': 'list[FormulaParameterOutputV1]',
        'properties': 'list[PropertyOutputV1]',
        'scoped_to': 'str',
        'status_message': 'str',
        'translation_key': 'str',
        'type': 'str',
        'value_unit_of_measure': 'str',
        'workbook_id': 'str'
    }

    attribute_map = {
        'cached': 'cached',
        'children': 'children',
        'created_by': 'createdBy',
        'datasource': 'datasource',
        'description': 'description',
        'effective_permissions': 'effectivePermissions',
        'fingerprint': 'fingerprint',
        'has_children': 'hasChildren',
        'id': 'id',
        'is_archived': 'isArchived',
        'is_redacted': 'isRedacted',
        'manually_added': 'manuallyAdded',
        'name': 'name',
        'parameters': 'parameters',
        'properties': 'properties',
        'scoped_to': 'scopedTo',
        'status_message': 'statusMessage',
        'translation_key': 'translationKey',
        'type': 'type',
        'value_unit_of_measure': 'valueUnitOfMeasure',
        'workbook_id': 'workbookId'
    }

    def __init__(self, cached=False, children=None, created_by=None, datasource=None, description=None, effective_permissions=None, fingerprint=None, has_children=False, id=None, is_archived=False, is_redacted=False, manually_added=None, name=None, parameters=None, properties=None, scoped_to=None, status_message=None, translation_key=None, type=None, value_unit_of_measure=None, workbook_id=None):
        """
        TreeItemOutputV1 - a model defined in Swagger
        """

        self._cached = None
        self._children = None
        self._created_by = None
        self._datasource = None
        self._description = None
        self._effective_permissions = None
        self._fingerprint = None
        self._has_children = None
        self._id = None
        self._is_archived = None
        self._is_redacted = None
        self._manually_added = None
        self._name = None
        self._parameters = None
        self._properties = None
        self._scoped_to = None
        self._status_message = None
        self._translation_key = None
        self._type = None
        self._value_unit_of_measure = None
        self._workbook_id = None

        if cached is not None:
          self.cached = cached
        if children is not None:
          self.children = children
        if created_by is not None:
          self.created_by = created_by
        if datasource is not None:
          self.datasource = datasource
        if description is not None:
          self.description = description
        if effective_permissions is not None:
          self.effective_permissions = effective_permissions
        if fingerprint is not None:
          self.fingerprint = fingerprint
        if has_children is not None:
          self.has_children = has_children
        if id is not None:
          self.id = id
        if is_archived is not None:
          self.is_archived = is_archived
        if is_redacted is not None:
          self.is_redacted = is_redacted
        if manually_added is not None:
          self.manually_added = manually_added
        if name is not None:
          self.name = name
        if parameters is not None:
          self.parameters = parameters
        if properties is not None:
          self.properties = properties
        if scoped_to is not None:
          self.scoped_to = scoped_to
        if status_message is not None:
          self.status_message = status_message
        if translation_key is not None:
          self.translation_key = translation_key
        if type is not None:
          self.type = type
        if value_unit_of_measure is not None:
          self.value_unit_of_measure = value_unit_of_measure
        if workbook_id is not None:
          self.workbook_id = workbook_id

    @property
    def cached(self):
        """
        Gets the cached of this TreeItemOutputV1.
        Indicates whether the item's persistent disk cache is enabled. Only present for items which support caching

        :return: The cached of this TreeItemOutputV1.
        :rtype: bool
        """
        return self._cached

    @cached.setter
    def cached(self, cached):
        """
        Sets the cached of this TreeItemOutputV1.
        Indicates whether the item's persistent disk cache is enabled. Only present for items which support caching

        :param cached: The cached of this TreeItemOutputV1.
        :type: bool
        """

        self._cached = cached

    @property
    def children(self):
        """
        Gets the children of this TreeItemOutputV1.
        Children of the tree node

        :return: The children of this TreeItemOutputV1.
        :rtype: list[TreeItemOutputV1]
        """
        return self._children

    @children.setter
    def children(self, children):
        """
        Sets the children of this TreeItemOutputV1.
        Children of the tree node

        :param children: The children of this TreeItemOutputV1.
        :type: list[TreeItemOutputV1]
        """

        self._children = children

    @property
    def created_by(self):
        """
        Gets the created_by of this TreeItemOutputV1.
        UUID of the user who created the asset tree

        :return: The created_by of this TreeItemOutputV1.
        :rtype: str
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """
        Sets the created_by of this TreeItemOutputV1.
        UUID of the user who created the asset tree

        :param created_by: The created_by of this TreeItemOutputV1.
        :type: str
        """

        self._created_by = created_by

    @property
    def datasource(self):
        """
        Gets the datasource of this TreeItemOutputV1.

        :return: The datasource of this TreeItemOutputV1.
        :rtype: DatasourcePreviewV1
        """
        return self._datasource

    @datasource.setter
    def datasource(self, datasource):
        """
        Sets the datasource of this TreeItemOutputV1.

        :param datasource: The datasource of this TreeItemOutputV1.
        :type: DatasourcePreviewV1
        """

        self._datasource = datasource

    @property
    def description(self):
        """
        Gets the description of this TreeItemOutputV1.
        Clarifying information or other plain language description of this item

        :return: The description of this TreeItemOutputV1.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this TreeItemOutputV1.
        Clarifying information or other plain language description of this item

        :param description: The description of this TreeItemOutputV1.
        :type: str
        """

        self._description = description

    @property
    def effective_permissions(self):
        """
        Gets the effective_permissions of this TreeItemOutputV1.

        :return: The effective_permissions of this TreeItemOutputV1.
        :rtype: PermissionsV1
        """
        return self._effective_permissions

    @effective_permissions.setter
    def effective_permissions(self, effective_permissions):
        """
        Sets the effective_permissions of this TreeItemOutputV1.

        :param effective_permissions: The effective_permissions of this TreeItemOutputV1.
        :type: PermissionsV1
        """

        self._effective_permissions = effective_permissions

    @property
    def fingerprint(self):
        """
        Gets the fingerprint of this TreeItemOutputV1.
        For stored items, an optional CDC (Change Data Capture) attribute.

        :return: The fingerprint of this TreeItemOutputV1.
        :rtype: str
        """
        return self._fingerprint

    @fingerprint.setter
    def fingerprint(self, fingerprint):
        """
        Sets the fingerprint of this TreeItemOutputV1.
        For stored items, an optional CDC (Change Data Capture) attribute.

        :param fingerprint: The fingerprint of this TreeItemOutputV1.
        :type: str
        """

        self._fingerprint = fingerprint

    @property
    def has_children(self):
        """
        Gets the has_children of this TreeItemOutputV1.
        True if this item has children

        :return: The has_children of this TreeItemOutputV1.
        :rtype: bool
        """
        return self._has_children

    @has_children.setter
    def has_children(self, has_children):
        """
        Sets the has_children of this TreeItemOutputV1.
        True if this item has children

        :param has_children: The has_children of this TreeItemOutputV1.
        :type: bool
        """

        self._has_children = has_children

    @property
    def id(self):
        """
        Gets the id of this TreeItemOutputV1.
        The ID that can be used to interact with the item

        :return: The id of this TreeItemOutputV1.
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """
        Sets the id of this TreeItemOutputV1.
        The ID that can be used to interact with the item

        :param id: The id of this TreeItemOutputV1.
        :type: str
        """
        if id is None:
            raise ValueError("Invalid value for `id`, must not be `None`")

        self._id = id

    @property
    def is_archived(self):
        """
        Gets the is_archived of this TreeItemOutputV1.
        Whether item is archived

        :return: The is_archived of this TreeItemOutputV1.
        :rtype: bool
        """
        return self._is_archived

    @is_archived.setter
    def is_archived(self, is_archived):
        """
        Sets the is_archived of this TreeItemOutputV1.
        Whether item is archived

        :param is_archived: The is_archived of this TreeItemOutputV1.
        :type: bool
        """

        self._is_archived = is_archived

    @property
    def is_redacted(self):
        """
        Gets the is_redacted of this TreeItemOutputV1.
        Whether item is redacted

        :return: The is_redacted of this TreeItemOutputV1.
        :rtype: bool
        """
        return self._is_redacted

    @is_redacted.setter
    def is_redacted(self, is_redacted):
        """
        Sets the is_redacted of this TreeItemOutputV1.
        Whether item is redacted

        :param is_redacted: The is_redacted of this TreeItemOutputV1.
        :type: bool
        """

        self._is_redacted = is_redacted

    @property
    def manually_added(self):
        """
        Gets the manually_added of this TreeItemOutputV1.
        UUID of the user who manually added the asset to the group

        :return: The manually_added of this TreeItemOutputV1.
        :rtype: str
        """
        return self._manually_added

    @manually_added.setter
    def manually_added(self, manually_added):
        """
        Sets the manually_added of this TreeItemOutputV1.
        UUID of the user who manually added the asset to the group

        :param manually_added: The manually_added of this TreeItemOutputV1.
        :type: str
        """

        self._manually_added = manually_added

    @property
    def name(self):
        """
        Gets the name of this TreeItemOutputV1.
        The human readable name

        :return: The name of this TreeItemOutputV1.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this TreeItemOutputV1.
        The human readable name

        :param name: The name of this TreeItemOutputV1.
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")

        self._name = name

    @property
    def parameters(self):
        """
        Gets the parameters of this TreeItemOutputV1.
        Formula Parameters used to create the formula

        :return: The parameters of this TreeItemOutputV1.
        :rtype: list[FormulaParameterOutputV1]
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """
        Sets the parameters of this TreeItemOutputV1.
        Formula Parameters used to create the formula

        :param parameters: The parameters of this TreeItemOutputV1.
        :type: list[FormulaParameterOutputV1]
        """

        self._parameters = parameters

    @property
    def properties(self):
        """
        Gets the properties of this TreeItemOutputV1.
        The list of the item's properties

        :return: The properties of this TreeItemOutputV1.
        :rtype: list[PropertyOutputV1]
        """
        return self._properties

    @properties.setter
    def properties(self, properties):
        """
        Sets the properties of this TreeItemOutputV1.
        The list of the item's properties

        :param properties: The properties of this TreeItemOutputV1.
        :type: list[PropertyOutputV1]
        """
        if properties is None:
            raise ValueError("Invalid value for `properties`, must not be `None`")

        self._properties = properties

    @property
    def scoped_to(self):
        """
        Gets the scoped_to of this TreeItemOutputV1.
        The ID of the workbook to which this item is scoped or null if it is in the global scope.

        :return: The scoped_to of this TreeItemOutputV1.
        :rtype: str
        """
        return self._scoped_to

    @scoped_to.setter
    def scoped_to(self, scoped_to):
        """
        Sets the scoped_to of this TreeItemOutputV1.
        The ID of the workbook to which this item is scoped or null if it is in the global scope.

        :param scoped_to: The scoped_to of this TreeItemOutputV1.
        :type: str
        """

        self._scoped_to = scoped_to

    @property
    def status_message(self):
        """
        Gets the status_message of this TreeItemOutputV1.
        A plain language status message with information about any issues that may have been encountered during an operation

        :return: The status_message of this TreeItemOutputV1.
        :rtype: str
        """
        return self._status_message

    @status_message.setter
    def status_message(self, status_message):
        """
        Sets the status_message of this TreeItemOutputV1.
        A plain language status message with information about any issues that may have been encountered during an operation

        :param status_message: The status_message of this TreeItemOutputV1.
        :type: str
        """

        self._status_message = status_message

    @property
    def translation_key(self):
        """
        Gets the translation_key of this TreeItemOutputV1.
        The item's translation key, if any

        :return: The translation_key of this TreeItemOutputV1.
        :rtype: str
        """
        return self._translation_key

    @translation_key.setter
    def translation_key(self, translation_key):
        """
        Sets the translation_key of this TreeItemOutputV1.
        The item's translation key, if any

        :param translation_key: The translation_key of this TreeItemOutputV1.
        :type: str
        """

        self._translation_key = translation_key

    @property
    def type(self):
        """
        Gets the type of this TreeItemOutputV1.
        The type of the item

        :return: The type of this TreeItemOutputV1.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Sets the type of this TreeItemOutputV1.
        The type of the item

        :param type: The type of this TreeItemOutputV1.
        :type: str
        """
        if type is None:
            raise ValueError("Invalid value for `type`, must not be `None`")

        self._type = type

    @property
    def value_unit_of_measure(self):
        """
        Gets the value_unit_of_measure of this TreeItemOutputV1.
        If item is a signal, the unit of measure for the values

        :return: The value_unit_of_measure of this TreeItemOutputV1.
        :rtype: str
        """
        return self._value_unit_of_measure

    @value_unit_of_measure.setter
    def value_unit_of_measure(self, value_unit_of_measure):
        """
        Sets the value_unit_of_measure of this TreeItemOutputV1.
        If item is a signal, the unit of measure for the values

        :param value_unit_of_measure: The value_unit_of_measure of this TreeItemOutputV1.
        :type: str
        """

        self._value_unit_of_measure = value_unit_of_measure

    @property
    def workbook_id(self):
        """
        Gets the workbook_id of this TreeItemOutputV1.
        If the item is a worksheet, the ID of its associated workbook

        :return: The workbook_id of this TreeItemOutputV1.
        :rtype: str
        """
        return self._workbook_id

    @workbook_id.setter
    def workbook_id(self, workbook_id):
        """
        Sets the workbook_id of this TreeItemOutputV1.
        If the item is a worksheet, the ID of its associated workbook

        :param workbook_id: The workbook_id of this TreeItemOutputV1.
        :type: str
        """

        self._workbook_id = workbook_id

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, TreeItemOutputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



from __future__ import annotations

import json
import os
import re
from typing import List, Dict, Union, Optional

import pandas as pd

from seeq import spy
from seeq.base import util
from seeq.sdk import *
from seeq.spy import _common, _metadata
from seeq.spy._errors import *
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy.workbooks._item import Item, ItemMap


class StoredOrCalculatedItem(Item):
    def __init__(self, definition=None, *, provenance=None):
        super().__init__(definition, provenance=provenance)

        self.scoped_to = None

    def _populate_asset_and_path(self, session: Session, item_id, status):
        trees_api = TreesApi(session.client)
        asset_tree_output = safely(
            lambda: trees_api.get_tree(id=item_id),
            action_description=f'get Tree information for {item_id}',
            status=status)  # type: AssetTreeOutputV1
        if asset_tree_output is None or asset_tree_output.item is None:
            return

        ancestors = [a.name for a in asset_tree_output.item.ancestors]

        _common.add_ancestors_to_definition(self.type, self.name, ancestors, self.definition, old_asset_format=False)

    def _pull(self, session: Session, item_id, status):
        super()._pull(session, item_id, status)

        if 'Path' not in self.definition:
            self._populate_asset_and_path(session, item_id, status)

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None,
             label=None, override_max_interp=False, scope_globals_to_workbook=False):
        raise SPyRuntimeError('Pushed called but StoredOrCalculatedItem.push() not overloaded')

    @staticmethod
    def _find_datasource_name(datasource_class, datasource_id, datasource_maps):
        for datasource_map in datasource_maps:
            if datasource_map['Datasource Class'] == datasource_class and \
                    datasource_map['Datasource ID'] == datasource_id:
                return datasource_map['Datasource Name']

        raise SPyRuntimeError('Could not find Datasource Class "%s" and Datasource ID "%s" in datasource maps' %
                              (datasource_class, datasource_id))

    @staticmethod
    def _execute_regex_map(old_definition, regex_map, regex_map_index, logging, *, allow_missing_properties=False):
        capture_groups = dict()
        for prop, regex in regex_map['Old'].items():
            if prop not in StoredItem.SEARCHABLE_PROPS:
                searchable_props_string = '\n'.join(StoredItem.SEARCHABLE_PROPS)
                raise SPyRuntimeError(
                    f'Datasource Map contains an unsearchable property "{prop}".'
                    f'Searchable properties:\n{searchable_props_string}')

            if prop not in old_definition:
                if allow_missing_properties:
                    continue
                else:
                    logging.append(f'- RegEx-Based Map {regex_map_index}: {prop} not defined for this item')
                    return None

            regex = util.pythonize_regex_capture_group_names(regex)
            match = re.fullmatch(regex, old_definition[prop])
            if not match:
                logging.append(
                    f'- RegEx-Based Map {regex_map_index}: {prop} "{old_definition[prop]}" does not match RegEx '
                    f'"{regex}"')
                return None

            capture_groups.update(match.groupdict())

        new_definition = dict()
        for prop, regex in regex_map['New'].items():
            new_definition[prop] = util.replace_tokens_in_regex(regex, capture_groups, escape=False)

        return new_definition

    def _lookup_item_via_datasource_map(self, session: Session, pushed_workbook_id: str, datasource_maps: List[Dict], *,
                                        only_override_maps: bool = False) -> \
            Optional[Union[ItemOutputV1, UserOutputV1, ItemPreviewV1]]:
        logging = list()
        items_api = ItemsApi(session.client)

        item: Optional[Union[ItemOutputV1, UserOutputV1, ItemPreviewV1]] = None

        # First, we process the "overrides". These are the cases where, even if the item with the ID exists in the
        # destination, we still want to map to something else. Useful for swapping datasources on the same server.
        overrides = [m for m in datasource_maps if _common.get(m, 'Override', default=False)]
        if len(overrides) > 0:
            if "File" in overrides[0]:
                logging.append(f'Using overrides from {os.path.dirname(overrides[0]["File"])}:')
            item = self._lookup_in_datasource_map(session, pushed_workbook_id, overrides, logging)

        if only_override_maps or item is not None:
            return item

        # Second, we just try to look the item up by its ID. This case will occur when the user pulls a workbook,
        # makes a change, and pushes it back.
        try:
            item = items_api.get_item_and_all_properties(id=self.id)  # type: ItemOutputV1
        except ApiException:
            logging.append(f'Item\'s ID {self.id} not found directly in target server')

        if item is not None:
            return item

        # Finally, we try to use the non-override maps to find the item. This case will occur mostly when
        # transferring workbooks between servers.
        non_overrides = [m for m in datasource_maps if not _common.get(m, 'Override', default=False)]
        if len(non_overrides) > 0:
            if "File" in non_overrides[0]:
                logging.append(f'Using non-overrides from {os.path.dirname(non_overrides[0]["File"])}:')
            item = self._lookup_in_datasource_map(session, pushed_workbook_id, non_overrides, logging)

        if item is None:
            logging_str = '\n'.join(logging)
            raise SPyDependencyNotFound(f'{self} not successfully mapped. More info:\n{logging_str}')

        return item

    def _lookup_in_datasource_map(self, session: Session, pushed_workbook_id, datasource_maps, logging) -> \
            Optional[Union[ItemOutputV1, UserOutputV1, ItemPreviewV1]]:
        item: Optional[Union[ItemOutputV1, UserOutputV1, ItemPreviewV1]] = None

        relevant_maps = list()
        for datasource_map in datasource_maps:
            if _common.get(self, 'Datasource Class') != _common.get(datasource_map, 'Datasource Class') or \
                    _common.get(self, 'Datasource ID') != _common.get(datasource_map, 'Datasource ID'):
                continue

            relevant_maps.append(datasource_map)
            if "File" in datasource_map:
                logging.append(f'- Used "{datasource_map["File"]}"')
            else:
                logging.append(f'- Used datasource map for Datasource Class "{datasource_map["Datasource Class"]}" '
                               f'and Datasource ID "{datasource_map["Datasource ID"]}"')

            if _common.DATASOURCE_MAP_ITEM_LEVEL_MAP_FILES in datasource_map:
                item = self._lookup_in_item_level_map_files(session, datasource_map, logging)

            if item is not None:
                break

            if _common.DATASOURCE_MAP_REGEX_BASED_MAPS in datasource_map:
                item = self._lookup_in_regex_based_map(session, datasource_map, datasource_maps, logging,
                                                       pushed_workbook_id)

            if item is not None:
                break

        if len(relevant_maps) == 0:
            logging.append(
                f'- No Datasource Maps found that match Datasource Class "{_common.get(self, "Datasource Class")}" '
                f'and Datasource ID "{_common.get(self, "Datasource ID")}"')

        return item

    def _lookup_in_item_level_map_files(self, session: Session, datasource_map, logging) -> Optional[ItemOutputV1]:
        items_api = ItemsApi(session.client)

        item: Optional[ItemOutputV1] = None
        item_level_map_files = datasource_map['Item-Level Map Files']
        if not isinstance(item_level_map_files, list):
            item_level_map_files = [item_level_map_files]

        if 'DataFrames' not in datasource_map:
            datasource_map['DataFrames'] = dict()

        df_map = datasource_map['DataFrames']

        for item_level_map_file in item_level_map_files:
            if item_level_map_file not in df_map:
                if not os.path.exists(item_level_map_file):
                    raise SPyValueError('Item-Level Map File "%s" does not exist' % item_level_map_file)

                df_map[item_level_map_file] = pd.read_csv(item_level_map_file)

            df = df_map[item_level_map_file]

            if 'Old ID' not in df.columns or 'New ID' not in df.columns:
                raise SPyValueError('Item-Level Map File must have "Old ID" and "New ID" columns')

            mapped_row = df[df['Old ID'] == self.id]

            if len(mapped_row) > 1:
                raise SPyValueError('Ambiguous map: %d rows in "%s" match "Old ID" == "%s"\n%s' % (
                    len(mapped_row), item_level_map_file, self.id, mapped_row))

            if len(mapped_row) == 0:
                logging.append('- Item-Level Map File "%s" did not match for "Old ID" value "%s"' % (
                    item_level_map_file, self.id))
                continue

            new_id = mapped_row.iloc[0]['New ID']

            item = items_api.get_item_and_all_properties(id=new_id)
            break

        return item

    def _lookup_in_regex_based_map(self, session: Session, datasource_map, datasource_maps, logging,
                                   pushed_workbook_id) -> Optional[Union[ItemOutputV1, UserOutputV1,
                                                                         IdentityPreviewV1]]:
        items_api = ItemsApi(session.client)

        item: Optional[Union[ItemOutputV1, UserOutputV1, IdentityPreviewV1]] = None
        new_definition = None
        regex_map_index = None
        for i in range(len(datasource_map['RegEx-Based Maps'])):
            regex_map_index = i
            regex_map = datasource_map['RegEx-Based Maps'][regex_map_index]
            old_definition = self.definition_dict

            if 'Type' not in old_definition:
                raise SPyValueError('"Type" property required in "Old" datasource map definitions')

            if 'Datasource Class' in old_definition and 'Datasource ID' in old_definition:
                old_definition['Datasource Name'] = \
                    StoredOrCalculatedItem._find_datasource_name(old_definition['Datasource Class'],
                                                                 old_definition['Datasource ID'],
                                                                 datasource_maps)

            new_definition = StoredOrCalculatedItem._execute_regex_map(
                old_definition, regex_map, regex_map_index, logging,
                allow_missing_properties=self.type.endswith('Datasource'))

            if new_definition is not None:
                break

        if new_definition is None:
            return None

        if 'Type' not in new_definition:
            raise SPyValueError('"Type" property required in "New" datasource map definitions')

        if 'Datasource Name' in new_definition:
            if 'Datasource ID' not in new_definition:
                if 'Datasource Class' not in new_definition:
                    raise SPyValueError('"Datasource Class" required with "Datasource Name" in map:\n%s' %
                                        json.dumps(new_definition))

                datasource_results = items_api.search_items(
                    # Note here that 'Name' supports RegEx but 'Datasource Class' doesn't. Users have to be careful in
                    # the Datasource Maps not to supply a RegEx. See Workbook._construct_default_datasource_maps().
                    filters=['Datasource Class==%s&&Name==/%s/' % (new_definition['Datasource Class'],
                                                                   new_definition['Datasource Name']),
                             '@includeUnsearchable'],
                    types=['Datasource'],
                    limit=2)  # type: ItemSearchPreviewPaginatedListV1

                items = datasource_results.items
                if len(items) > 1:
                    # Try filtering out any archived items first
                    items = [i for i in items if not i.is_archived]

                if len(items) > 1:
                    multiple_items_str = "\n".join([i.id for i in items])
                    raise SPyRuntimeError(
                        f'Multiple datasources found that match "{new_definition["Datasource Name"]}":\n'
                        f'{multiple_items_str}'
                    )
                elif len(items) == 0:
                    raise SPyRuntimeError(
                        f'No datasource found that matches "{new_definition["Datasource Name"]}"'
                    )

                new_datasource = items[0]  # type: ItemSearchPreviewV1
                try:
                    new_definition['Datasource ID'] = items_api.get_property(
                        id=new_datasource.id, property_name='Datasource ID').value
                except ApiException:
                    raise SPyRuntimeError(f'Could not get Datasource ID for '
                                          f'"{new_definition["Datasource Name"]}" {new_datasource.id}')

            del new_definition['Datasource Name']

        # Simplify the types so that the search will find stored or calculated signals/conditions etc. This
        # helps when mapping, for example, an AF tree to a Data Lab tree.
        for t in ['Signal', 'Condition', 'Scalar', 'Datasource']:
            if t in new_definition['Type']:
                new_definition['Type'] = t

        if new_definition['Type'] not in ['User', 'UserGroup']:
            query_df = pd.DataFrame([new_definition])
            from seeq.spy import _search
            if 'Path' in new_definition:
                # If we're matching based on path, unfortunately we will not be able to match against archived items
                # due to a limitation of the API. See CRAB-18439.
                search_df = _search.search(query_df, workbook=pushed_workbook_id, recursive=False,
                                           ignore_unindexed_properties=False, session=session, quiet=True)
            else:
                search_df = _search.search(query_df, workbook=pushed_workbook_id, include_archived=True,
                                           ignore_unindexed_properties=False, session=session, quiet=True)

            # If several things are returned, but some of them are archived, then filter out the archived ones.
            # Tested by spy.workbooks.tests.test_push.test_datasource_map_by_name()
            if len(search_df) > 1 and 'Archived' in search_df.columns:
                search_df = search_df[~search_df['Archived']]

            if len(search_df) == 0:
                logging.append(f'- RegEx-Based Map {regex_map_index}: No match for search:\n'
                               f'{new_definition}')
            elif len(search_df) > 1:
                logging.append(f'- RegEx-Based Map {regex_map_index}: Multiple matches for search:\n'
                               f'{new_definition}\n'
                               f'{str(search_df[["ID", "Name"]])}')
            else:
                item = items_api.get_item_and_all_properties(
                    id=search_df.iloc[0]['ID'])  # type: ItemOutputV1
        else:
            if new_definition['Type'] == 'User':
                try:
                    users_api = UsersApi(session.client)
                    item = users_api.get_user_from_username(
                        auth_datasource_class=new_definition['Datasource Class'],
                        auth_datasource_id=new_definition['Datasource ID'],
                        username=new_definition['Username'])  # type: UserOutputV1
                except ApiException:
                    # Fall through, item not found
                    pass
            else:
                user_groups_api = UserGroupsApi(session.client)
                user_groups = user_groups_api.get_user_groups()
                for user_group in user_groups.items:  # type: IdentityPreviewV1
                    if user_group.name == new_definition['Name']:
                        item = user_group
                        break

        return item


class StoredItem(StoredOrCalculatedItem):
    SEARCHABLE_PROPS = ['Datasource Class', 'Datasource ID', 'Datasource Name', 'Data ID',
                        'Type', 'Name', 'Description', 'Username', 'Path', 'Asset']

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        item = self._lookup_item_via_datasource_map(session, pushed_workbook_id, datasource_maps)

        item_map[self.id] = item.id

        if item.type not in ['User', 'UserGroup']:
            # We need to exclude Example Data, because it is set explicitly by the connector
            datasource_class_prop = Item._property_output_from_item_output(item, 'Datasource Class')
            datasource_id_prop = Item._property_output_from_item_output(item, 'Datasource ID')
            is_example_data = (datasource_class_prop and datasource_class_prop.value == 'Time Series CSV Files' and
                               datasource_id_prop and datasource_id_prop.value == 'Example Data')

            if override_max_interp and item.type == 'StoredSignal' and not is_example_data and \
                    'Maximum Interpolation' in self:
                src_max_interp = Item._property_input_from_scalar_str(self['Maximum Interpolation'])
                dst_max_interp_prop = Item._property_output_from_item_output(item, 'Maximum Interpolation')
                if dst_max_interp_prop:
                    dst_max_interp = Item._property_input_from_scalar_str(dst_max_interp_prop.value)
                    if src_max_interp.unit_of_measure != dst_max_interp.unit_of_measure or \
                            src_max_interp.value != dst_max_interp.value:
                        items_api = ItemsApi(session.client)
                        items_api.set_property(id=item.id,
                                               property_name='Override Maximum Interpolation',
                                               body=src_max_interp)

        return item


class Asset(StoredItem):
    def __init__(self, definition=None, *, provenance=None):
        super().__init__(definition, provenance=provenance)

        if self.provenance == Item.LOAD:
            if self.definition.get('Old Asset Format', True):
                if self.type == 'Asset':
                    if self.definition.get('Path') in [None, '']:
                        self.definition['Path'] = self.definition['Asset']
                    else:
                        self.definition['Path'] += ' >> ' + self.definition['Asset']
                        self.definition['Asset'] = self.definition['Name']

        self.definition['Old Asset Format'] = False


class Datasource(StoredItem):
    @staticmethod
    def from_datasource_output(datasource_output: Union[DatasourceOutputV1, DatasourcePreviewV1]):
        return Datasource({
            'ID': datasource_output.id,
            'Name': datasource_output.name,
            'Datasource Class': datasource_output.datasource_class,
            'Datasource ID': datasource_output.datasource_id,
            'Archived': datasource_output.is_archived
        })


class TableDatasource(Datasource):
    pass


class CalculatedItem(StoredOrCalculatedItem):
    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        # This is a key piece of datasource mapping logic. CalculatedItems that are in trees are more like "source
        # data" in the context of workbooks/worksheets. In other words, when we push them, we want to map in existing
        # items rather than creating a standalone CalculatedItem that isn't a part of a tree, has no associated asset
        # and cannot be swapped.
        is_standalone_item = _common.get(self, 'Asset') is None

        # So if it's in a tree, we process the datasource maps just like we would if it were a StoredItem. If it's
        # *not* in a tree, then we only process datasource maps that have been supplied in a datasource_map_folder
        # and are therefore treated as "overrides". CalculatedItems that are associated with the particular analysis
        # of a workbook should be created/overwritten as part of the push, and since they won't exist in a tree,
        # we purposefully don't attempt to do any mapping and we fall through to the CalculatedItem's create/update
        # code.
        item = self._lookup_item_via_datasource_map(session, pushed_workbook_id, datasource_maps,
                                                    only_override_maps=is_standalone_item)

        if item:
            item_map[self.id] = item.id

        return item

    def _find(self, session: Session, label, datasource_output, pushed_workbook_id):
        item_output = self.find_me(session, label, datasource_output)

        if item_output is None and self.provenance == Item.CONSTRUCTOR:
            item_output = self._find_by_name(session, pushed_workbook_id)

        return item_output

    def _find_by_name(self, session: Session, workbook_id):
        items_api = ItemsApi(session.client)

        _filters = ['Name==%s' % self.name,
                    '@includeUnsearchable']

        search_results = items_api.search_items(
            filters=_filters,
            scope=[workbook_id],
            types=[self.type],
            offset=0,
            limit=1)  # type: ItemSearchPreviewPaginatedListV1

        if len(search_results.items) == 0:
            return None

        return search_results.items[0]

    def _create_calculated_item_input(self, clazz, item_map: ItemMap, scoped_to, scope_globals_to_workbook):
        item_input = clazz()
        item_input.name = self.definition['Name']
        if 'Description' in self.definition:
            item_input.description = self.definition['Description']
        item_input.formula = Item.formula_string_from_list(self.definition['Formula'])

        parameters = list()
        if 'Formula Parameters' in self.definition:
            for parameter_name, parameter_id in self.definition['Formula Parameters'].items():
                if parameter_id not in item_map:
                    raise SPyDependencyNotFound(f'{self} formula parameter {parameter_name}={parameter_id} not found')

                parameters.append('%s=%s' % (parameter_name, item_map[parameter_id]))

        for _attr_name in ['formula_parameters', 'parameters']:
            if hasattr(item_input, _attr_name):
                setattr(item_input, _attr_name, parameters)

        if 'Number Format' in self.definition:
            item_input.number_format = self.definition['Number Format']

        if scope_globals_to_workbook or _common.present(self.definition, 'Scoped To'):
            item_input.scoped_to = scoped_to

        return item_input

    def _set_ui_config(self, session: Session, _id):
        if 'UIConfig' in self.definition:
            try:
                items_api = ItemsApi(session.client)
                items_api.set_property(id=_id, property_name='UIConfig',
                                       body=PropertyInputV1(value=json.dumps(self.definition['UIConfig'])))
            except ValueError:  # ValueError includes JSONDecodeError, but ApiException will not be caught
                pass


class StoredSignal(StoredItem):
    pass


class CalculatedSignal(CalculatedItem):
    def _pull(self, session: Session, item_id, status):
        super()._pull(session, item_id, status)
        # Note: The caller of this function is in charge of API safety
        signals_api = SignalsApi(session.client)
        signal_output = signals_api.get_signal(id=item_id)  # type: SignalOutputV1
        self._set_formula_based_item_properties(signal_output)

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        session = Session.validate(session)
        # Note: The caller of this function is in charge of API safety
        item = super().push(session, datasource_maps, datasource_output, pushed_workbook_id=pushed_workbook_id,
                            item_map=item_map, label=label, override_max_interp=override_max_interp,
                            scope_globals_to_workbook=scope_globals_to_workbook)

        if item:
            return item

        signals_api = SignalsApi(session.client)

        item_output = self._find(session, label, datasource_output, pushed_workbook_id)

        signal_input = self._create_calculated_item_input(SignalInputV1, item_map,
                                                          pushed_workbook_id,
                                                          scope_globals_to_workbook)  # type: SignalInputV1

        if item_output is None:
            signal_output = signals_api.put_signal_by_data_id(datasource_class=datasource_output.datasource_class,
                                                              datasource_id=datasource_output.datasource_id,
                                                              data_id=self._construct_data_id(label),
                                                              body=signal_input)  # type: SignalOutputV1
        else:
            signal_output = signals_api.put_signal(id=item_output.id,
                                                   body=signal_input)  # type: SignalOutputV1

        self._set_ui_config(session, signal_output.id)

        item_map[self.id] = signal_output.id

        return signal_output


class StoredCondition(StoredItem):
    pass


class CalculatedCondition(CalculatedItem):
    def _pull(self, session: Session, item_id, status):
        super()._pull(session, item_id, status)
        # Note: The caller of this function is in charge of API safety
        conditions_api = ConditionsApi(session.client)
        condition_output = conditions_api.get_condition(id=item_id)  # type: ConditionOutputV1
        self._set_formula_based_item_properties(condition_output)

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        # Note: The caller of this function is in charge of API safety
        item = super().push(session, datasource_maps, datasource_output, pushed_workbook_id=pushed_workbook_id,
                            item_map=item_map, label=label, override_max_interp=override_max_interp,
                            scope_globals_to_workbook=scope_globals_to_workbook)

        if item:
            return item

        conditions_api = ConditionsApi(session.client)

        item_output = self._find(session, label, datasource_output, pushed_workbook_id)

        condition_update_input = self._create_calculated_item_input(
            ConditionUpdateInputV1, item_map,
            pushed_workbook_id, scope_globals_to_workbook)  # type: ConditionUpdateInputV1
        condition_update_input.replace_capsule_properties = True

        if item_output is None:
            condition_update_input.datasource_class = datasource_output.datasource_class
            condition_update_input.datasource_id = datasource_output.datasource_id
            condition_update_input.data_id = self._construct_data_id(label)
        else:
            # There is no easy way to update a Condition by its ID, so we have to retrieve its data id triplet
            condition_output = conditions_api.get_condition(id=item_output.id)  # type: ConditionOutputV1
            condition_update_input.datasource_class = condition_output.datasource_class
            condition_update_input.datasource_id = condition_output.datasource_id
            condition_update_input.data_id = condition_output.data_id

        item_batch_output = conditions_api.put_conditions(
            body=ConditionBatchInputV1(conditions=[condition_update_input]))  # type: ItemBatchOutputV1

        item_update_output = item_batch_output.item_updates[0]  # type: ItemUpdateOutputV1
        if item_update_output.error_message is not None:
            raise SPyRuntimeError('Could not push condition "%s": %s' %
                                  (self.definition['Name'], item_update_output.error_message))

        self._set_ui_config(session, item_update_output.item.id)

        item_map[self.id] = item_update_output.item.id

        return item_update_output.item


class CalculatedScalar(CalculatedItem):
    def _pull(self, session: Session, item_id, status):
        super()._pull(session, item_id, status)
        # Note: The caller of this function is in charge of API safety
        scalars_api = ScalarsApi(session.client)
        calculated_item_output = scalars_api.get_scalar(id=item_id)  # type: CalculatedItemOutputV1
        self._set_formula_based_item_properties(calculated_item_output)

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        # Note: The caller of this function is in charge of API safety
        item = super().push(session, datasource_maps, datasource_output, pushed_workbook_id=pushed_workbook_id,
                            item_map=item_map, label=label, override_max_interp=override_max_interp,
                            scope_globals_to_workbook=scope_globals_to_workbook)

        if item:
            return item

        scalars_api = ScalarsApi(session.client)
        items_api = ItemsApi(session.client)

        item_output = self._find(session, label, datasource_output, pushed_workbook_id)

        scalar_input = self._create_calculated_item_input(
            ScalarInputV1, item_map,
            pushed_workbook_id, scope_globals_to_workbook)  # type: ScalarInputV1

        if item_output is None:
            datasource_class = datasource_output.datasource_class
            datasource_id = datasource_output.datasource_id
            scalar_input.data_id = self._construct_data_id(label)

            if (spy.utils.is_sdk_module_version_at_least(61, 1, 4) and not spy.utils.is_sdk_module_version_at_least(
                    62)) or spy.utils.is_sdk_module_version_at_least(62, 0, 1):
                body = PutScalarsInputV1(datasource_class=datasource_class,
                                         datasource_id=datasource_id,
                                         scalars=[scalar_input],
                                         force_calculated_scalars=True)
            else:
                body = PutScalarsInputV1(datasource_class=datasource_class,
                                         datasource_id=datasource_id,
                                         scalars=[scalar_input])

            item_batch_output = scalars_api.put_scalars(body=body)  # type: ItemBatchOutputV1

            item_update_output = item_batch_output.item_updates[0]  # type: ItemUpdateOutputV1
            if item_update_output.error_message is not None:
                raise SPyRuntimeError('Could not push scalar "%s": %s' %
                                      (self.definition['Name'], item_update_output.error_message))

            self._set_ui_config(session, item_update_output.item.id)
            item_map[self.id] = item_update_output.item.id

            return item_update_output.item

        else:
            # There is no easy way to update a Scalar by its ID, so we have to use ItemsApi
            items_api.set_formula(id=item_output.id, body=FormulaUpdateInputV1(formula=scalar_input.formula,
                                                                               parameters=scalar_input.parameters))
            if scalar_input.scoped_to:
                items_api.set_scope(id=item_output.id, workbook_id=scalar_input.scoped_to)
            else:
                items_api.set_scope(id=item_output.id)
            props = [ScalarPropertyV1(name='Name', value=scalar_input.name)]
            if 'Definition' in self:
                props.append(ScalarPropertyV1(name='Description', value=scalar_input.description))
            if 'Number Format' in self:
                props.append(ScalarPropertyV1(name='Number Format', value=scalar_input.number_format))
            items_api.set_properties(id=item_output.id, body=props)
            self._set_ui_config(session, item_output.id)
            item_map[self.id] = item_output.id

            return item_output


class Chart(CalculatedItem):
    def _pull(self, session: Session, item_id, status):
        super()._pull(session, item_id, status)
        # Note: The caller of this function is in charge of API safety
        formulas_api = FormulasApi(session.client)
        calculated_item_output = formulas_api.get_function(id=item_id)  # type: CalculatedItemOutputV1

        self._set_formula_based_item_properties(calculated_item_output)

        if 'FormulaParameters' in self.definition:
            # Charts have these superfluous properties
            del self.definition['FormulaParameters']

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        # Note: The caller of this function is in charge of API safety
        item = super().push(session, datasource_maps, datasource_output, pushed_workbook_id=pushed_workbook_id,
                            item_map=item_map, label=label, override_max_interp=override_max_interp,
                            scope_globals_to_workbook=scope_globals_to_workbook)

        if item:
            return item

        formulas_api = FormulasApi(session.client)
        items_api = ItemsApi(session.client)

        item_output = self._find(session, label, datasource_output, pushed_workbook_id)

        function_input = FunctionInputV1()
        function_input.name = self.definition['Name']
        function_input.type = self.definition['Type']
        function_input.formula = Item.formula_string_from_list(self.definition['Formula'])
        function_input.parameters = list()
        for parameter_name, parameter_id in self.definition['Formula Parameters'].items():
            if _common.is_guid(parameter_id):
                if parameter_id not in item_map:
                    raise SPyDependencyNotFound(
                        f'{self} formula parameter {parameter_name} ({parameter_id}) not pushed')

                function_input.parameters.append(FormulaParameterInputV1(name=parameter_name,
                                                                         id=item_map[parameter_id]))
            else:
                function_input.parameters.append(FormulaParameterInputV1(name=parameter_name,
                                                                         formula=parameter_id,
                                                                         unbound=True))

        if 'Description' in self.definition:
            function_input.description = self.definition['Description']

        function_input.scoped_to = pushed_workbook_id
        function_input.data_id = self._construct_data_id(label)

        if item_output is None:
            calculated_item_output = formulas_api.create_function(body=function_input)  # type: CalculatedItemOutputV1

            items_api.set_properties(
                id=calculated_item_output.id,
                body=[ScalarPropertyV1(name='Datasource Class', value=datasource_output.datasource_class),
                      ScalarPropertyV1(name='Datasource ID', value=datasource_output.datasource_id),
                      ScalarPropertyV1(name='Data ID', value=function_input.data_id)])
        else:
            calculated_item_output = formulas_api.update_function(id=item_output.id,
                                                                  body=function_input)  # type: CalculatedItemOutputV1

        self._set_ui_config(session, calculated_item_output.id)

        item_map[self.id] = calculated_item_output.id

        return calculated_item_output


class ThresholdMetric(CalculatedItem):
    def _pull(self, session: Session, item_id, status):
        super()._pull(session, item_id, status)
        # Note: The caller of this function is in charge of API safety
        formula_parameters = _metadata.formula_parameters_dict_from_threshold_metric(session, item_id)

        # These properties come through in the GET /items/{id} call, and for clarity's sake we remove them
        for ugly_duplicate_property in ['AggregationFunction', 'BoundingConditionMaximumDuration']:
            if ugly_duplicate_property in self.definition:
                del self.definition[ugly_duplicate_property]

        self.definition['Formula'] = '<ThresholdMetric>'
        self.definition['Formula Parameters'] = formula_parameters

    def push(self, session: Session, datasource_maps, datasource_output, *, pushed_workbook_id=None,
             item_map: ItemMap = None, label=None, override_max_interp=False, scope_globals_to_workbook=False):
        # Note: The caller of this function is in charge of API safety
        item = super().push(session, datasource_maps, datasource_output, pushed_workbook_id=pushed_workbook_id,
                            item_map=item_map, label=label, override_max_interp=override_max_interp,
                            scope_globals_to_workbook=scope_globals_to_workbook)

        if item:
            return item

        items_api = ItemsApi(session.client)
        metrics_api = MetricsApi(session.client)

        parameters = self['Formula Parameters']

        new_item = _metadata.threshold_metric_input_from_formula_parameters(parameters, item_object=self,
                                                                            item_map=item_map)

        new_item.name = self.name
        new_item.scoped_to = pushed_workbook_id
        new_item.number_format = _common.get(parameters, 'Number Format')

        item_output = self._find(session, label, datasource_output, pushed_workbook_id)

        while True:
            try:
                if item_output is None:
                    threshold_metric_output = metrics_api.create_threshold_metric(
                        body=new_item)  # type: ThresholdMetricOutputV1

                    items_api.set_properties(
                        id=threshold_metric_output.id,
                        body=[ScalarPropertyV1(name='Datasource Class', value=datasource_output.datasource_class),
                              ScalarPropertyV1(name='Datasource ID', value=datasource_output.datasource_id),
                              ScalarPropertyV1(name='Data ID', value=self._construct_data_id(label))])
                else:
                    threshold_metric_output = metrics_api.put_threshold_metric(
                        id=item_output.id,
                        body=new_item)  # type: ThresholdMetricOutputV1

                break

            except ApiException as e:
                # We have to handle a case where a condition on which a metric depends has been changed from bounded
                # to unbounded. In the UI, it automatically fills in the default of 40h when you edit such a metric,
                # so we do roughly the same thing here. This is tested by test_push.test_bad_metric().
                exception_text = _common.format_exception(e)
                if 'Maximum Capsule Duration for Bounding Condition must be provided' in exception_text:
                    new_item.bounding_condition_maximum_duration = '40h'
                else:
                    raise

        self._set_ui_config(session, threshold_metric_output.id)

        item_map[self.id] = threshold_metric_output.id

        return threshold_metric_output



from __future__ import annotations

from typing import Optional

import seeq.spy._errors as errors
from seeq.sdk import ApiClient
from seeq.sdk.configuration import Configuration
from seeq.sdk.models import UserOutputV1
from seeq.spy import acl
from seeq.spy import addons
from seeq.spy import assets
from seeq.spy import docs
from seeq.spy import jobs
from seeq.spy import notifications
from seeq.spy import utils
from seeq.spy import widgets
from seeq.spy import workbooks
from seeq.spy._common import PATH_ROOT, DEFAULT_WORKBOOK_PATH, GLOBALS_ONLY, GLOBALS_AND_ALL_WORKBOOKS, \
    INHERIT_FROM_WORKBOOK
from seeq.spy._login import login, logout
from seeq.spy._plot import plot
from seeq.spy._pull import pull
from seeq.spy._push import push
from seeq.spy._search import search
from seeq.spy._session import Session, Options
from seeq.spy._status import Status
from seeq.spy._swap import swap
from seeq.spy._upgrade import upgrade

# noinspection DuplicatedCode
session: Session = Session(client_configuration=Configuration())
"""
The default session used by SPy functions that interact with Seeq Server
"""

options: Options = session.options
"""
Equivalent to `spy.session.options`
"""

client: Optional[ApiClient] = None
"""
Equivalent to `spy.session.client`
"""

user: Optional[UserOutputV1] = None
"""
Equivalent to `spy.session.user`
"""

server_version: Optional[str] = None
"""
Equivalent to `spy.session.server_version`
"""

__all__ = ['acl', 'addons', 'assets', 'docs', 'workbooks', 'widgets', 'login', 'logout', 'plot', 'pull', 'push',
           'search', 'swap', 'PATH_ROOT', 'DEFAULT_WORKBOOK_PATH', 'GLOBALS_ONLY', 'GLOBALS_AND_ALL_WORKBOOKS',
           'INHERIT_FROM_WORKBOOK', 'Session', 'Status', 'options', 'session', 'client', 'user',
           'server_version', 'jobs', 'notifications', 'upgrade', 'utils', 'errors']

__version__ = '%d.%d' % (int('189'), int('11'))



from __future__ import annotations

import dataclasses
import fnmatch
import functools
import itertools
import re
from collections import deque, defaultdict
from enum import Enum
from typing import Dict, List, Union, Tuple, Callable, Iterable, Set, Hashable, Optional

import numpy as np
import pandas as pd

from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy.assets._trees import _path
from seeq.spy.assets._trees._pandas import KeyedDataFrame

Pattern = Union[None, 'TreeNode', pd.DataFrame, List['Pattern'], dict, pd.Series, str, int]


@dataclasses.dataclass(frozen=True)
class TreeNode:
    """
    This is a helper class used to navigate the internal DataFrame of a spy.assets.Tree object in a more traditional
    object-oriented manner. Each object of this type corresponds to a row in said DataFrame.

    Currently this class is used for:
    - Resolving formula parameters referenced by path/name in the tree
    - Querying items in the tree by their name/path/depth/ID/other properties

    The KeyedDataFrame class exists solely so that the TreeNode.of() static constructor can have its results cached.
    This ensures that whenever a tree has its internal DataFrame state modified, a TreeNode version of its state is
    saved as well via the TreeNode.of() cache, and queries on its items can be made quickly without traversing the
    DataFrame again.

    NOTE: equality and hashing is determined entirely by the `index` attribute. Do not compare TreeNodes that
    originate from different DataFrames
    """
    id: str = dataclasses.field(init=True, repr=False, compare=False)
    name: str = dataclasses.field(init=True, repr=True, compare=False)
    type: str = dataclasses.field(init=True, repr=False, compare=False)
    depth: int = dataclasses.field(init=True, repr=False, compare=False)
    referenced_id: str = dataclasses.field(init=True, repr=False, compare=False)
    children: Tuple[TreeNode] = dataclasses.field(init=True, repr=False, compare=False)
    path: str = dataclasses.field(init=False, repr=True, compare=False)
    index: int = dataclasses.field(init=True, repr=False, compare=True)
    size: int = dataclasses.field(init=True, repr=False, compare=False)
    parent: TreeNode = dataclasses.field(init=False, repr=False, compare=False)
    root: TreeNode = dataclasses.field(init=False, repr=False, compare=False)

    @staticmethod
    def of(df: KeyedDataFrame) -> TreeNode:
        return TreeNode._of(df.reset_index().to_dict(orient='records'))

    @staticmethod
    def _of(records: List[dict], set_as_root=True, offset=0) -> TreeNode:
        root = records[offset]
        root_depth = root['Depth']
        size: int

        def children_generator() -> Iterable[TreeNode]:
            i = offset + 1
            while i < len(records) and records[i]['Depth'] > root_depth:
                child = TreeNode._of(records, set_as_root=False, offset=i)
                yield child
                i += child.size
            nonlocal size
            size = i - offset

        children = tuple(children_generator())

        this = TreeNode(id=_common.get(root, 'ID'),
                        name=_common.get(root, 'Name'),
                        type=_common.get(root, 'Type'),
                        depth=_common.get(root, 'Depth'),
                        referenced_id=_common.get(root, 'Referenced ID'),
                        index=root['index'],
                        size=size,
                        children=children)
        for child in children:
            object.__setattr__(child, 'parent', this)  # This bypasses immutability so we can late-initialize fields
        if set_as_root:
            object.__setattr__(this, 'parent', None)
            this._set_path_and_root_rec(path='', root=this)
        return this

    def _set_path_and_root_rec(self, path: str, root: TreeNode):
        object.__setattr__(self, 'path', path)
        object.__setattr__(self, 'root', root)
        for child in self.children:
            child._set_path_and_root_rec(self.full_path, root)

    @property
    def full_path(self) -> str:
        return f'{self.path} >> {self.name}' if self.path else self.name

    def __iter__(self) -> Iterable[TreeNode]:
        yield self
        for child in self.children:
            yield from child

    @property
    def ancestors(self) -> Iterable[TreeNode]:
        node = self
        while node != self.root:
            node = node.parent
            yield node
        yield node

    def matches(self, pattern: Pattern) -> bool:
        if pattern is None:
            return self == self.root
        if pd.api.types.is_scalar(pattern) and pd.isnull(pattern):
            # This case handles when the user only gives the 'Parent' column for some children, or gives a parent
            #  string that uses column values that aren't valid for some rows.
            return False
        if isinstance(pattern, TreeNode):
            return self == pattern
        if isinstance(pattern, pd.DataFrame):
            return self.matches(pattern.to_dict(orient='records'))
        if isinstance(pattern, list):
            return any(self.matches(sub_pattern) for sub_pattern in pattern)
        if pd.api.types.is_dict_like(pattern):
            if _common.present(pattern, 'ID'):
                return pattern['ID'] == self.id or pattern['ID'] == self.referenced_id
            if _common.present(pattern, 'Name'):
                return (not _common.present(pattern, 'Path') or _path.determine_path(pattern) == self.path) \
                    and pattern['Name'] == self.name
            return False
        if isinstance(pattern, str):
            if _common.is_guid(pattern):
                return pattern.upper() == self.id
            else:
                return self.matches_path(_common.path_string_to_list(pattern))
        if isinstance(pattern, int):
            return pattern == self.depth
        return False

    def matches_path(self, path_list: List[str]) -> bool:
        name_regex = exact_or_glob_or_regex(path_list.pop())
        if not cached_regex_eval(name_regex, self.name):
            return False
        if len(path_list) == 0:
            return True
        if not self.parent:
            return False
        return self.parent.matches_path(path_list)

    def is_name_match(self, name: str) -> bool:
        name_regex = exact_or_glob_or_regex(name)
        return bool(cached_regex_eval(name_regex, self.name))

    def resolve_reference(self, reference: str) -> TreeNode:
        reference_name = f'{self.path} >> {reference}'

        def reference_at_node(node: TreeNode, relative_path: str, full_name: str, check_self=False) -> TreeNode:
            _result = None
            for i, match in enumerate(node._relative_matches(relative_path, check_self=check_self)):
                if i == 0:
                    if match.type == 'Asset':
                        raise SPyRuntimeError(f'Formula parameter "{full_name}" is an asset. Formula parameters '
                                              f'must be conditions, scalars, or signals.')
                    _result = match
                else:
                    raise SPyRuntimeError(f'Formula parameter "{full_name}" matches multiple items in tree.')
            return _result

        relative = reference_at_node(self.parent, reference, reference_name, check_self=False)
        if relative is None:
            if reference.startswith(self.root.name):  # Assume that no one will try to wildcard-match the root
                absolute = reference_at_node(self.root, reference, reference, check_self=True)
                if absolute is None:
                    raise SPyRuntimeError(
                        f'Formula parameter is invalid, missing, or has been removed from tree: "{reference}".')
                else:
                    return absolute
            else:
                raise SPyRuntimeError(
                    f'Formula parameter is invalid, missing, or has been removed from tree: "{reference_name}".')
        else:
            return relative

    def resolve_references(self, reference: str) -> List[TreeNode]:
        reference_name = f'{self.path} >> {reference}'

        def references_at_node(node: TreeNode, relative_path: str, full_name: str, check_self=False) -> TreeNode:
            for match in node._relative_matches(relative_path, check_self=check_self):
                if match.type == 'Asset':
                    raise SPyRuntimeError(f'Formula parameter "{full_name}" is an asset. Formula parameters '
                                          f'must be conditions, scalars, or signals.')
                yield match

        relative = list(references_at_node(self.parent, reference, reference_name, check_self=False))
        if len(relative) == 0 and reference.startswith(self.root.name):
            absolute = list(references_at_node(self.root, reference, reference, check_self=True))
            if len(absolute) != 0:
                return absolute
        return relative

    def _relative_matches(self, relative_path: Union[str, List[str]], check_self=False, offset=0) -> Iterable[TreeNode]:
        components = _common.path_string_to_list(relative_path) if isinstance(relative_path, str) else relative_path
        if offset < len(components):
            if components[offset] == '..':
                if not check_self and self.parent is not None:
                    yield from self.parent._relative_matches(components, check_self=False, offset=offset + 1)
            else:
                if check_self:
                    if self.is_name_match(components[offset]):
                        if offset == len(components) - 1:
                            yield self
                        elif components[offset + 1] == '..':
                            if self.parent is not None:
                                yield from self.parent._relative_matches(components, check_self=False,
                                                                         offset=offset + 2)
                        else:
                            for child in self.children:
                                yield from child._relative_matches(components, check_self=True, offset=offset + 1)
                else:
                    for child in self.children:
                        yield from child._relative_matches(components, check_self=True, offset=offset)


class Query:
    """
    This non-user-facing class is used to select items from a tree's internal DataFrame.
    """

    class Operations(Enum):
        AND = 0
        OR = 1
        REPLACE = 2
        DIFFERENCE = 3

    class Matchers(Enum):
        CHILDREN = 0
        DESCENDANTS = 1
        SIBLINGS = 2
        PARENTS = 3
        ANCESTORS = 4

    MATCHING_FUNCTIONS: Dict[Query.Matchers, Callable[[TreeNode], Iterable[TreeNode]]] = {
        Matchers.CHILDREN: lambda node: node.children,
        Matchers.DESCENDANTS: lambda node: itertools.chain(*node.children),
        Matchers.SIBLINGS: lambda node: (sibling for sibling in node.parent.children if sibling != node),
        Matchers.PARENTS: lambda node: (node.parent,),
        Matchers.ANCESTORS: lambda node: node.ancestors,
    }

    def __init__(self, df: KeyedDataFrame):
        self.df = df
        self.tree = TreeNode.of(self.df)
        self.filters = deque()
        self.matched = set(self.tree)

    def copy(self) -> Query:
        new = Query(self.df)
        new.filter = self.filters.copy()
        if len(self.matched) != self.tree.size:
            new.matched = self.matched.copy()
        return new

    def matches(self, pattern: Pattern | Query | Callable) -> Query:
        self.filters.append((Query.Operations.AND, pattern))
        return self

    def has_type(self, item_type: str) -> Query:
        self.filters.append((Query.Operations.AND, lambda node: item_type in node.type))
        return self

    def exclude(self, pattern: Pattern) -> Query:
        self.filters.append((Query.Operations.DIFFERENCE, pattern))
        return self

    def with_children(self) -> Query:
        self.filters.append((Query.Operations.OR, Query.Matchers.CHILDREN))
        return self

    def children(self) -> Query:
        self.filters.append((Query.Operations.REPLACE, Query.Matchers.CHILDREN))
        return self

    def with_descendants(self) -> Query:
        self.filters.append((Query.Operations.OR, Query.Matchers.DESCENDANTS))
        return self

    def descendants(self) -> Query:
        self.filters.append((Query.Operations.REPLACE, Query.Matchers.DESCENDANTS))
        return self

    def with_siblings(self) -> Query:
        self.filters.append((Query.Operations.OR, Query.Matchers.SIBLINGS))
        return self

    def siblings(self) -> Query:
        self.filters.append((Query.Operations.REPLACE, Query.Matchers.SIBLINGS))
        return self

    def with_parents(self) -> Query:
        self.filters.append((Query.Operations.OR, Query.Matchers.PARENTS))
        return self

    def parents(self) -> Query:
        self.filters.append((Query.Operations.REPLACE, Query.Matchers.PARENTS))
        return self

    def with_ancestors(self) -> Query:
        self.filters.append((Query.Operations.OR, Query.Matchers.ANCESTORS))
        return self

    def ancestors(self) -> Query:
        self.filters.append((Query.Operations.REPLACE, Query.Matchers.ANCESTORS))
        return self

    def _execute(self):
        while self.filters:
            self._filter(*self.filters.popleft())

    def _filter(self, operation: Operations, matcher: Union[Query, Matchers, Callable, Pattern]):
        if isinstance(matcher, Query.Matchers):
            matching_function = Query.MATCHING_FUNCTIONS[matcher]
            new_matches = set(itertools.chain.from_iterable(matching_function(node) for node in self.matched))
        elif isinstance(matcher, Query):
            matcher._execute()
            new_matches = matcher.matched
        elif callable(matcher):
            nodes_to_check = self.matched if operation in (Query.Operations.AND,
                                                           Query.Operations.DIFFERENCE) else self.tree
            new_matches = set(node for node in nodes_to_check if matcher(node))
        else:
            nodes_to_check = self.matched if operation in (Query.Operations.AND,
                                                           Query.Operations.DIFFERENCE) else self.tree
            new_matches = set(node for node in nodes_to_check if node.matches(matcher))
        if operation == Query.Operations.AND:
            self.matched &= new_matches
        elif operation == Query.Operations.OR:
            self.matched |= new_matches
        elif operation == Query.Operations.REPLACE:
            self.matched = new_matches
        elif operation == Query.Operations.DIFFERENCE:
            self.matched -= new_matches

    def get_filtered_rows(self) -> KeyedDataFrame:
        self._execute()
        out_df = self.df.loc[self.get_mask()]
        return out_df

    def get_mask(self) -> np.ndarray:
        self._execute()
        return self.df.index.isin(set(node.index for node in self.matched))

    def get_node_list(self, sort=False) -> List[TreeNode]:
        self._execute()
        if sort:
            return [node for node in self.tree if node in self.matched]
        else:
            return list(self.matched)

    def get_node_set(self) -> Set[TreeNode]:
        self._execute()
        return self.matched

    def get_distinct_subtree_roots(self) -> Iterable[TreeNode]:
        self._execute()
        it = iter(self.tree)
        node = next(it, None)
        while node is not None:
            if node in self.matched:
                yield node
                node = next(itertools.islice(it, node.size - 1, node.size), None)
            else:
                node = next(it, None)

    def multimatch(self, pattern_series: pd.Series) -> Iterable[TreeNode, Set[int]]:
        """
        Make multiple queries at once using a pd.Series object containing multiple Patterns to match.
        Returns an generator, each element of which specifies a TreeNode that matched at least one query
        along with the indices of all given queries that it matched.

        Used by Tree.insert() when the user is inserting a DataFrame with a 'Parent' column
        """
        pattern_indices = defaultdict(set)
        unhashable_patterns = list()
        for index, pattern in pattern_series.items():
            if pd.isnull(pattern):
                continue
            if isinstance(pattern, Hashable):
                pattern_indices[pattern].add(index)
            else:
                unhashable_patterns.append((pattern, {index}))

        node_indices = defaultdict(set)
        for pattern, indices in itertools.chain(pattern_indices.items(), unhashable_patterns):
            for node in self.copy().matches(pattern).get_node_set():
                node_indices[node].update(indices)
        return node_indices.items()


def is_column_value_query(s):
    if not isinstance(s, str):
        return False
    if re.search(r'{{.*}.*}', s):
        return True
    return False


def fill_column_values(row, query: str = None, query_column=None):
    """
    Fills a column values query with actual column values from a row in a dataframe. Returns the output string
    """
    if pd.isnull(query):
        if query_column not in row:
            return np.nan
        query = row[query_column]
        if pd.isnull(query):
            return np.nan

    def _fill_column_value(col_val_query_match: re.Match):
        col_val_query = col_val_query_match[1]
        col, extract_pattern = re.fullmatch(r'{(.*?)}(.*)', col_val_query).groups(default='')
        if not _common.present(row, col):
            raise SPyValueError('Not a match')
        value = str(row[col])
        if extract_pattern == '':
            return value

        # Match against a glob pattern first, then try regex
        for pattern in (glob_with_capture_groups_to_regex(extract_pattern), extract_pattern):
            try:
                extraction = re.fullmatch(pattern, value)
                if extraction:
                    if len(extraction.groups()) != 0:
                        return extraction[1]
                    else:
                        return extraction[0]
            except re.error:
                # There may be a compilation error if the input wasn't intended to be interpreted as regex
                continue
        raise SPyValueError('Not a match')

    try:
        return re.sub(r'{({.*?}.*?)}', _fill_column_value, query)
    except SPyValueError:
        return np.nan


def fnmatch_translate_from_py3_7_8(pat):
    """
    This code is copied from Python 3.7.8 source code. This function changed significantly in 3.10 and is
    incompatible with the fixups that we do in glob_with_capture_groups_to_regex() to hack in a "capture group"
    facility to the glob.
    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i + 1
        if c == '*':
            res = res + '.*'
        elif c == '?':
            res = res + '.'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j + 1
            if j < n and pat[j] == ']':
                j = j + 1
            while j < n and pat[j] != ']':
                j = j + 1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j]
                if '--' not in stuff:
                    stuff = stuff.replace('\\', r'\\')
                else:
                    chunks = []
                    k = i + 2 if pat[i] == '!' else i + 1
                    while True:
                        k = pat.find('-', k, j)
                        if k < 0:
                            break
                        chunks.append(pat[i:k])
                        i = k + 1
                        k = k + 3
                    chunks.append(pat[i:j])
                    # Escape backslashes and hyphens for set difference (--).
                    # Hyphens that create ranges shouldn't be escaped.
                    stuff = '-'.join(s.replace('\\', r'\\').replace('-', r'\-')
                                     for s in chunks)
                # Escape set operations (&&, ~~ and ||).
                stuff = re.sub(r'([&~|])', r'\\\1', stuff)
                i = j + 1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] in ('^', '['):
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return r'(?s:%s)\Z' % res


def glob_with_capture_groups_to_regex(glob):
    """
    Converts a glob to a regex, but does not escape parentheses, so that the glob can be written with capture groups
    """
    return re.sub(r'\\([()])', r'\1', fnmatch_translate_from_py3_7_8(glob))


@functools.lru_cache(maxsize=2048)
def exact_or_glob_or_regex(pat: str) -> re.Pattern:
    try:
        return re.compile('(?i)' + '(' + ')|('.join([re.escape(pat), fnmatch.translate(pat), pat]) + ')')
    except re.error:
        return re.compile('(?i)' + '(' + ')|('.join([re.escape(pat), fnmatch.translate(pat)]) + ')')


@functools.lru_cache(maxsize=2048)
def cached_regex_eval(pattern: re.Pattern, value: str) -> Optional[re.Match]:
    return pattern.fullmatch(value)
